@@ -697,7 +697,7 @@ last_component (path)
 
     last = strrchr (path, '/');
 
-    if (last)
+    if (last && (last != path))
         return last + 1;
     else
         return path;
