@@ -15,32 +15,6 @@
  * or the configure script directly.  Sorry.
  */
 
-/*
- * CVS provides the most features when used in conjunction with the Version-5
- * release of RCS.  Thus, it is the default.  This also assumes that GNU diff
- * Version-1.15 is being used as well -- you will have to configure your RCS
- * V5 release separately to make this the case. If you do not have RCS V5 and
- * GNU diff V1.15, comment out this define. You should not try mixing and
- * matching other combinations of these tools.
- */
-#ifndef HAVE_RCS5
-#define	HAVE_RCS5
-#endif
-
-/*
- * If, before installing this version of CVS, you were running RCS V4 AND you
- * are installing this CVS and RCS V5 and GNU diff 1.15 all at the same time,
- * you should turn on the following define.  It only exists to try to do
- * reasonable things with your existing checked out files when you upgrade to
- * RCS V5, since the keyword expansion formats have changed with RCS V5.
- * 
- * If you already have been running with RCS5, or haven't been running with CVS
- * yet at all, or are sticking with RCS V4 for now, leave the commented out.
- */
-#ifndef HAD_RCS4
-/* #define	HAD_RCS4 */
-#endif
-
 /*
  * For portability and heterogeneity reasons, CVS is shipped by default using
  * my own text-file version of the ndbm database library in the src/myndbm.c
@@ -51,34 +25,6 @@
 #define	MY_NDBM
 #endif
 
-/*
- * The "diff" program to execute when creating patch output.  This "diff"
- * must support the "-c" option for context diffing.  Specify a full
- * pathname if your site wants to use a particular diff.  If you are
- * using the GNU version of diff (version 1.15 or later), this should
- * be "diff -a".
- * 
- * NOTE: this program is only used for the ``patch'' sub-command (and
- * for ``update'' if you are using the server).  The other commands
- * use rcsdiff which will use whatever version of diff was specified
- * when rcsdiff was built on your system.
- */
-
-#ifndef DIFF
-#define	DIFF	"@gdiff_path@"
-#endif
-
-/*
- * The "grep" program to execute when checking to see if a merged file had
- * any conflicts.  This "grep" must support the "-s" option and a standard
- * regular expression as an argument.  Specify a full pathname if your site
- * wants to use a particular grep.
- */
-
-#ifndef GREP
-#define GREP "@ggrep_path@"
-#endif
-
 /*
  * The "patch" program to run when using the CVS server and accepting
  * patches across the network.  Specify a full pathname if your site
@@ -88,31 +34,12 @@
 #define PATCH_PROGRAM	"patch"
 #endif
 
-/*
- * By default, RCS programs are executed with the shell or through execlp(),
- * so the user's PATH environment variable is searched.  If you'd like to
- * bind all RCS programs to a certain directory (perhaps one not in most
- * people's PATH) then set the default in RCSBIN_DFLT.  Note that setting
- * this here will cause all RCS programs to be executed from this directory,
- * unless the user overrides the default with the RCSBIN environment variable
- * or the "-b" option to CVS.
- * 
- * This define should be either the empty string ("") or a full pathname to the
- * directory containing all the installed programs from the RCS distribution.
- */
-#ifndef RCSBIN_DFLT
-#define	RCSBIN_DFLT	""
-#endif
+/* Directory used for storing temporary files, if not overridden by
+   environment variables or the -T global option.  There should be little
+   need to change this (-T is a better mechanism if you need to use a
+   different directory for temporary files).
 
-/*
- * The password-authenticating server creates a temporary checkout of
- * the affected files.  The variable TMPDIR_DFLT (or even better, the
- * command-line option "-T" in the line for CVS in /etc/inetd.conf)
- * can be used to specify the used directory.  This directory will
- * also be used for other temporary files.
- *
- * I have no idea what the right default for this is on the Mac.
- */
+   I have no idea what the right default for this is on the Mac.  */
 #ifndef TMPDIR_DFLT
 #define	TMPDIR_DFLT	"/tmp"
 #endif
@@ -247,12 +174,6 @@
  */
 #undef SETXID_SUPPORT
 
-/*
- * "cvs login" is under construction.  Don't define this unless you're
- * testing it, in which case you're me and you already know that.
- */
-/* #define CVS_LOGIN */
-
 /* End of CVS configuration section */
 
 /*
@@ -262,8 +183,3 @@
 #ifndef STDC_HEADERS
 extern void exit ();
 #endif
-
-#ifndef getwd
-extern char *getwd ();
-#endif
-
