@@ -17,6 +17,7 @@
  *  X		is a single character showing the type of event:
  *		T	"Tag" cmd.
  *		O	"Checkout" cmd.
+ *              E       "Export" cmd.
  *		F	"Release" cmd.
  *		W	"Update" cmd - No User file, Remove from Entries file.
  *		U	"Update" cmd - File was checked out over User file.
@@ -40,18 +41,18 @@
  *		command was typed.
  *		T	"A" --> New Tag, "D" --> Delete Tag
  *			Otherwise it is the Tag or Date to modify.
- *		O,F	A "" (null field)
+ *		O,F,E	A "" (null field)
  *
  *  rev(s)	Revision number or tag.
  *		T	The Tag to apply.
- *		O	The Tag or Date, if specified, else "" (null field).
+ *		O,E	The Tag or Date, if specified, else "" (null field).
  *		F	"" (null field)
  *		W	The Tag or Date, if specified, else "" (null field).
  *		U	The Revision checked out over the User file.
  *		G,C	The Revision(s) involved in merge.
  *		M,A,R	RCS Revision affected.
  *
- *  argument	The module (for [TOUF]) or file (for [WUGCMAR]) affected.
+ *  argument	The module (for [TOEUF]) or file (for [WUGCMAR]) affected.
  *
  *
  *** Report categories: "User" and "Since" modifiers apply to all reports.
@@ -59,7 +60,7 @@
  *
  *   Extract list of record types
  *
- *	-e, -x [TOFWUGCMAR]
+ *	-e, -x [TOEFWUGCMAR]
  *
  *		Extracted records are simply printed, No analysis is performed.
  *		All "field" modifiers apply.  -e chooses all types.
@@ -177,6 +178,7 @@
  */
 
 #include "cvs.h"
+#include "savecwd.h"
 
 static struct hrec
 {
@@ -206,7 +208,7 @@ static void save_file PROTO((char *dir, char *name, char *module));
 static void save_module PROTO((char *module));
 static void save_user PROTO((char *name));
 
-#define ALL_REC_TYPES "TOFWUCGMAR"
+#define ALL_REC_TYPES "TOEFWUCGMAR"
 #define USER_INCREMENT	2
 #define FILE_INCREMENT	128
 #define MODULE_INCREMENT 5
@@ -228,21 +230,28 @@ static short repos_sort;
 static short file_sort;
 static short module_sort;
 
-#ifdef HAVE_RCS5
 static short tz_local;
 static time_t tz_seconds_east_of_GMT;
 static char *tz_name = "+0000";
-#else
-static char tz_name[] = "LT";
-#endif
 
+/* -r, -t, or -b options, malloc'd.  These are "" if the option in
+   question is not specified or is overridden by another option.  The
+   main reason for using "" rather than NULL is historical.  Together
+   with since_date, these are a mutually exclusive set; one overrides the
+   others.  */
+static char *since_rev;
+static char *since_tag;
+static char *backto;
+/* -D option, or 0 if not specified.  */
 static time_t since_date;
-static char since_rev[20];	/* Maxrev ~= 99.99.99.999 */
-static char since_tag[64];
+
 static struct hrec *last_since_tag;
-static char backto[128];
 static struct hrec *last_backto;
-static char rec_types[20];
+
+/* Record types to look for, malloc'd.  Probably could be statically
+   allocated, but only if we wanted to check for duplicates more than
+   we do.  */
+static char *rec_types;
 
 static int hrec_count;
 static int hrec_max;
@@ -265,6 +274,12 @@ static int mod_count;		/* Number of elements used */
 
 static char *histfile;		/* Ptr to the history file name */
 
+/* This is pretty unclear.  First of all, separating "flags" vs.
+   "options" (I think the distinction is that "options" take arguments)
+   is nonstandard, and not something we do elsewhere in CVS.  Second of
+   all, what does "reports" mean?  I think it means that you can only
+   supply one of those options, but "reports" hardly has that meaning in
+   a self-explanatory way.  */
 static const char *const history_usg[] =
 {
     "Usage: %s %s [-report] [-flags] [-options args] [files...]\n\n",
@@ -273,7 +288,7 @@ static const char *const history_usg[] =
     "        -c              Committed (Modified) files\n",
     "        -o              Checked out modules\n",
     "        -m <module>     Look for specified module (repeatable)\n",
-    "        -x [TOFWUCGMAR] Extract by record type\n",
+    "        -x [TOEFWUCGMAR] Extract by record type\n",
     "   Flags:\n",
     "        -a              All users (Default is self)\n",
     "        -e              Everything (same as -x, but all record types)\n",
@@ -367,13 +382,17 @@ history (argc, argv)
     char **argv;
 {
     int i, c;
-    char fname[PATH_MAX];
+    char *fname;
 
     if (argc == -1)
 	usage (history_usg);
 
-    optind = 1;
-    while ((c = getopt (argc, argv, "Tacelow?D:b:f:m:n:p:r:t:u:x:X:z:")) != -1)
+    since_rev = xstrdup ("");
+    since_tag = xstrdup ("");
+    backto = xstrdup ("");
+    rec_types = xstrdup ("");
+    optind = 0;
+    while ((c = getopt (argc, argv, "+Tacelow?D:b:f:m:n:p:r:t:u:x:X:z:")) != -1)
     {
 	switch (c)
 	{
@@ -391,7 +410,8 @@ history (argc, argv)
 	    case 'e':
 		report_count++;
 		extract++;
-		(void) strcpy (rec_types, ALL_REC_TYPES);
+		free (rec_types);
+		rec_types = xstrdup (ALL_REC_TYPES);
 		break;
 	    case 'l':			/* Find Last file record */
 		last_entry = 1;
@@ -421,13 +441,8 @@ history (argc, argv)
 		    *since_rev = *since_tag = '\0';
 		    since_date = 0;
 		}
-		if (strlen (optarg) >= sizeof (backto))
-		{
-		    error (0, 0, "backto truncated to %d bytes",
-			   (int) sizeof (backto) - 1);
-		    optarg[sizeof (backto) - 1] = '\0';
-		}
-		(void) strcpy (backto, optarg);
+		free (backto);
+		backto = xstrdup (optarg);
 		break;
 	    case 'f':			/* For specified file */
 		save_file ("", optarg, (char *) NULL);
@@ -448,7 +463,8 @@ history (argc, argv)
 		    *since_tag = *backto = '\0';
 		    since_date = 0;
 		}
-		(void) strcpy (since_rev, optarg);
+		free (since_rev);
+		since_rev = xstrdup (optarg);
 		break;
 	    case 't':			/* Since specified Tag/Rev */
 		if (since_date || *since_rev || *backto)
@@ -457,7 +473,8 @@ history (argc, argv)
 		    *since_rev = *backto = '\0';
 		    since_date = 0;
 		}
-		(void) strcpy (since_tag, optarg);	/* tag */
+		free (since_tag);
+		since_tag = xstrdup (optarg);
 		break;
 	    case 'u':			/* For specified username */
 		save_user (optarg);
@@ -472,12 +489,10 @@ history (argc, argv)
 			if (!strchr (ALL_REC_TYPES, *cp))
 			    error (1, 0, "%c is not a valid report type", *cp);
 		}
-		(void) strcpy (rec_types, optarg);
+		free (rec_types);
+		rec_types = xstrdup (optarg);
 		break;
 	    case 'z':
-#ifndef HAVE_RCS5
-		error (0, 0, "-z not supported with RCS 4");
-#else
 		tz_local = 
 		    (optarg[0] == 'l' || optarg[0] == 'L')
 		    && (optarg[1] == 't' || optarg[1] == 'T')
@@ -508,7 +523,6 @@ history (argc, argv)
 			tz_name = optarg;
 		    }
 		}
-#endif
 		break;
 	    case '?':
 	    default:
@@ -588,7 +602,10 @@ history (argc, argv)
     if (tag_report)
     {
 	if (!strchr (rec_types, 'T'))
+	{
+	    rec_types = xrealloc (rec_types, strlen (rec_types) + 5);
 	    (void) strcat (rec_types, "T");
+	}
     }
     else if (extract)
     {
@@ -597,7 +614,8 @@ history (argc, argv)
     }
     else if (modified)
     {
-	(void) strcpy (rec_types, "MAR");
+	free (rec_types);
+	rec_types = xstrdup ("MAR");
 	/*
 	 * If the user has not specified a date oriented flag ("Since"), sort
 	 * by Repository/file before date.  Default is "just" date.
@@ -616,7 +634,8 @@ history (argc, argv)
     }
     else if (module_report)
     {
-	(void) strcpy (rec_types, last_entry ? "OMAR" : ALL_REC_TYPES);
+	free (rec_types);
+	rec_types = xstrdup (last_entry ? "OMAR" : ALL_REC_TYPES);
 	module_sort++;
 	repos_sort++;
 	file_sort++;
@@ -625,7 +644,8 @@ history (argc, argv)
     else
 	/* Must be "checkout" or default */
     {
-	(void) strcpy (rec_types, "OF");
+	free (rec_types);
+	rec_types = xstrdup ("OF");
 	/* See comments in "modified" above */
 	if (!last_entry && user_list)
 	    user_sort++;
@@ -639,7 +659,10 @@ history (argc, argv)
 
     /* If we're looking back to a Tag value, must consider "Tag" records */
     if (*since_tag && !strchr (rec_types, 'T'))
+    {
+	rec_types = xrealloc (rec_types, strlen (rec_types) + 5);
 	(void) strcat (rec_types, "T");
+    }
 
     argc -= c;
     argv += c;
@@ -647,14 +670,23 @@ history (argc, argv)
 	save_file ("", argv[i], (char *) NULL);
 
     if (histfile)
-	(void) strcpy (fname, histfile);
+	fname = xstrdup (histfile);
     else
+    {
+	fname = xmalloc (strlen (CVSroot_directory) + sizeof (CVSROOTADM)
+			 + sizeof (CVSROOTADM_HISTORY) + 10);
 	(void) sprintf (fname, "%s/%s/%s", CVSroot_directory,
 			CVSROOTADM, CVSROOTADM_HISTORY);
+    }
 
     read_hrecs (fname);
     qsort ((PTR) hrec_head, hrec_count, sizeof (struct hrec), sort_order);
     report_hrecs ();
+    free (fname);
+    free (since_rev);
+    free (since_tag);
+    free (backto);
+    free (rec_types);
 
     return (0);
 }
@@ -667,7 +699,8 @@ history_write (type, update_dir, revs, name, repository)
     char *name;
     char *repository;
 {
-    char fname[PATH_MAX], workdir[PATH_MAX], homedir[PATH_MAX];
+    char *fname;
+    char *workdir;
     char *username = getcaller ();
     int fd;
     char *line;
@@ -678,6 +711,8 @@ history_write (type, update_dir, revs, name, repository)
 
     if (logoff)			/* History is turned off by cmd line switch */
 	return;
+    fname = xmalloc (strlen (CVSroot_directory) + sizeof (CVSROOTADM)
+		     + sizeof (CVSROOTADM_HISTORY) + 10);
     (void) sprintf (fname, "%s/%s/%s", CVSroot_directory,
 		    CVSROOTADM, CVSROOTADM_HISTORY);
 
@@ -685,7 +720,7 @@ history_write (type, update_dir, revs, name, repository)
     if (!isfile (fname))
     {
 	logoff = 1;
-	return;
+	goto out;
     }
 
     if (trace)
@@ -696,7 +731,7 @@ history_write (type, update_dir, revs, name, repository)
 	fprintf (stderr, "-> fopen(%s,a)\n", fname);
 #endif
     if (noexec)
-	return;
+	goto out;
     fd = CVS_OPEN (fname, O_WRONLY | O_APPEND | O_CREAT | OPEN_BINARY, 0666);
     if (fd < 0)
 	error (1, errno, "cannot open history file: %s", fname);
@@ -721,13 +756,21 @@ history_write (type, update_dir, revs, name, repository)
 	    else
 	    {
 		/* Try harder to find a "homedir" */
-		if (!getwd (workdir))
-		    error (1, errno, "can't getwd in history");
+		struct saved_cwd cwd;
+		char *homedir;
+
+		if (save_cwd (&cwd))
+		    error_exit ();
+
 		if ( CVS_CHDIR (pwdir) < 0)
 		    error (1, errno, "can't chdir(%s)", pwdir);
-		if (!getwd (homedir))
+		homedir = xgetwd ();
+		if (homedir == NULL)
 		    error (1, errno, "can't getwd in %s", pwdir);
-		(void) CVS_CHDIR (workdir);
+
+		if (restore_cwd (&cwd, NULL))
+		    error_exit ();
+		free_cwd (&cwd);
 
 		i = strlen (homedir);
 		if (!strncmp (CurDir, homedir, i))
@@ -735,6 +778,7 @@ history_write (type, update_dir, revs, name, repository)
 		    PrCurDir += i;	/* Point to '/' separator */
 		    tilde = "~";
 		}
+		free (homedir);
 	    }
 	}
     }
@@ -749,6 +793,8 @@ history_write (type, update_dir, revs, name, repository)
     else
 	update_dir = "";
 
+    workdir = xmalloc (strlen (tilde) + strlen (PrCurDir) + strlen (slash)
+		       + strlen (update_dir) + 10);
     (void) sprintf (workdir, "%s%s%s%s", tilde, PrCurDir, slash, update_dir);
 
     /*
@@ -822,6 +868,9 @@ history_write (type, update_dir, revs, name, repository)
     free (line);
     if (close (fd) != 0)
 	error (1, errno, "cannot close history file: %s", fname);
+    free (workdir);
+ out:
+    free (fname);
 }
 
 /*
@@ -969,7 +1018,7 @@ fill_hrec (line, hr)
     NEXT_BAR (repos);
     NEXT_BAR (rev);
     hr->idx = idx++;
-    if (strchr ("FOT", *(hr->type)))
+    if (strchr ("FOET", *(hr->type)))
 	hr->mod = line;
 
     NEXT_BAR (file);	/* This returns ptr to next line or final '\0' */
@@ -1219,7 +1268,7 @@ select_hrec (hr)
      */
     if (!strchr (rec_types, *(hr->type)))
 	return (0);
-    if (!strchr ("TFO", *(hr->type)))	/* Don't bother with "file" if "TFO" */
+    if (!strchr ("TFOE", *(hr->type)))	/* Don't bother with "file" if "TFOE" */
     {
 	if (file_list)			/* If file_list is null, accept all */
 	{
@@ -1231,7 +1280,7 @@ select_hrec (hr)
 		 *    the concatenation of the repository and file from hrec.
 		 * 3. Else compare the file_list entry against the hrec file.
 		 */
-		char cmpfile[PATH_MAX];
+		char *cmpfile = NULL;
 
 		if (*(cp = fl->l_file) == '*')
 		{
@@ -1247,8 +1296,12 @@ select_hrec (hr)
 		{
 		    if (strchr (cp, '/'))
 		    {
-			(void) sprintf (cp2 = cmpfile, "%s/%s",
+			cmpfile = xmalloc (strlen (hr->repos)
+					   + strlen (hr->file)
+					   + 10);
+			(void) sprintf (cmpfile, "%s/%s",
 					hr->repos, hr->file);
+			cp2 = cmpfile;
 		    }
 		    else
 		    {
@@ -1261,6 +1314,8 @@ select_hrec (hr)
 			hr->mod = fl->l_module;
 			break;
 		    }
+		    if (cmpfile != NULL)
+			free (cmpfile);
 		}
 	    }
 	    if (!count)
@@ -1319,7 +1374,7 @@ report_hrecs ()
     hr++;
     for (count = hrec_count; count--; lr = hr, hr++)
     {
-	char repos[PATH_MAX];
+	char *repos;
 
 	if (!count)
 	    hr = NULL;
@@ -1327,7 +1382,7 @@ report_hrecs ()
 	    continue;
 
 	ty = *(lr->type);
-	(void) strcpy (repos, lr->repos);
+	repos = xstrdup (lr->repos);
 	if ((cp = strrchr (repos, '/')) != NULL)
 	{
 	    if (lr->mod && !strcmp (++cp, lr->mod))
@@ -1345,6 +1400,7 @@ report_hrecs ()
 	    rev_len = i;
 	if (lr->mod && (i = strlen (lr->mod)) > mod_len)
 	    mod_len = i;
+	free (repos);
     }
 
     /* Walk through hrec array setting "lr" (Last Record) to each element.
@@ -1358,7 +1414,8 @@ report_hrecs ()
      */
     for (lr = hrec_head, hr = (lr + 1); hrec_count--; lr = hr, hr++)
     {
-	char workdir[PATH_MAX], repos[PATH_MAX];
+	char *workdir;
+	char *repos;
 
 	if (!hrec_count)
 	    hr = NULL;
@@ -1366,19 +1423,19 @@ report_hrecs ()
 	    continue;
 
 	ty = *(lr->type);
-#ifdef HAVE_RCS5
 	if (!tz_local)
 	{
 	    time_t t = lr->date + tz_seconds_east_of_GMT;
 	    tm = gmtime (&t);
 	}
 	else
-#endif
-	tm = localtime (&(lr->date));
+	    tm = localtime (&(lr->date));
+
 	(void) printf ("%c %02d/%02d %02d:%02d %s %-*s", ty, tm->tm_mon + 1,
 		  tm->tm_mday, tm->tm_hour, tm->tm_min, tz_name,
 		  user_len, lr->user);
 
+	workdir = xmalloc (strlen (lr->dir) + strlen (lr->end) + 10);
 	(void) sprintf (workdir, "%s%s", lr->dir, lr->end);
 	if ((cp = strrchr (workdir, '/')) != NULL)
 	{
@@ -1387,6 +1444,7 @@ report_hrecs ()
 		(void) strcpy (cp, "*");
 	    }
 	}
+	repos = xmalloc (strlen (lr->repos) + 10);
 	(void) strcpy (repos, lr->repos);
 	if ((cp = strrchr (repos, '/')) != NULL)
 	{
@@ -1406,6 +1464,7 @@ report_hrecs ()
 		    (void) printf (" {%s}", workdir);
 		break;
 	    case 'F':
+	    case 'E':
 	    case 'O':
 		if (lr->rev && *(lr->rev))
 		    (void) printf (" [%s]", lr->rev);
@@ -1429,6 +1488,8 @@ report_hrecs ()
 		break;
 	}
 	(void) putchar ('\n');
+	free (workdir);
+	free (repos);
     }
 }
 
