@@ -3,7 +3,7 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  * 
  * "import" checks in the vendor release located in the current directory into
  * the CVS source repository.  The CVS vendor branch support is utilized.
@@ -18,13 +18,11 @@
 
 #include "cvs.h"
 #include "savecwd.h"
+#include <assert.h>
 
 #define	FILE_HOLDER	".#cvsxxx"
 
 static char *get_comment PROTO((char *user));
-static int add_rcs_file PROTO((char *message, char *rcs, char *user, char *vtag,
-		         int targc, char *targv[]));
-static int expand_at_signs PROTO((char *buf, off_t size, FILE *fp));
 static int add_rev PROTO((char *message, RCSNode *rcs, char *vfile,
 			  char *vers));
 static int add_tags PROTO((RCSNode *rcs, char *vfile, char *vtag, int targc,
@@ -39,10 +37,10 @@ static int update_rcs_file PROTO((char *message, char *vfile, char *vtag, int ta
 static void add_log PROTO((int ch, char *fname));
 
 static int repos_len;
-static char vhead[50];
-static char vbranch[50];
+static char *vhead;
+static char *vbranch;
 static FILE *logfp;
-static char repository[PATH_MAX];
+static char *repository;
 static int conflicts;
 static int use_file_modtime;
 static char *keyword_opt = NULL;
@@ -57,6 +55,7 @@ static const char *const import_usage[] =
     "\t-b bra\tVendor branch id.\n",
     "\t-m msg\tLog message.\n",
     "\t-W spec\tWrappers specification line.\n",
+    "(Specify the --help global option for a list of other help options)\n",
     NULL
 };
 
@@ -79,9 +78,9 @@ import (argc, argv)
     ign_setup ();
     wrap_setup ();
 
-    (void) strcpy (vbranch, CVSBRANCH);
-    optind = 1;
-    while ((c = getopt (argc, argv, "Qqdb:m:I:k:W:")) != -1)
+    vbranch = xstrdup (CVSBRANCH);
+    optind = 0;
+    while ((c = getopt (argc, argv, "+Qqdb:m:I:k:W:")) != -1)
     {
 	switch (c)
 	{
@@ -100,13 +99,14 @@ import (argc, argv)
 		use_file_modtime = 1;
 		break;
 	    case 'b':
-		(void) strcpy (vbranch, optarg);
+		free (vbranch);
+		vbranch = xstrdup (optarg);
 		break;
 	    case 'm':
 #ifdef FORCE_USE_EDITOR
-		use_editor = TRUE;
+		use_editor = 1;
 #else
-		use_editor = FALSE;
+		use_editor = 0;
 #endif
 		message = xstrdup(optarg);
 		break;
@@ -135,7 +135,14 @@ import (argc, argv)
 	usage (import_usage);
 
     for (i = 1; i < argc; i++)		/* check the tags for validity */
+    {
+	int j;
+
 	RCS_check_tag (argv[i]);
+	for (j = 1; j < i; j++)
+	    if (strcmp (argv[j], argv[i]) == 0)
+		error (1, 0, "tag `%s' was specified more than once", argv[i]);
+    }
 
     /* XXX - this should be a module, not just a pathname */
     if (! isabsolute (argv[0]))
@@ -146,11 +153,14 @@ import (argc, argv)
 	    error (1, 0, "Set it or specify the '-d' option to %s.",
 		   program_name);
 	}
+	repository = xmalloc (strlen (CVSroot_directory) + strlen (argv[0])
+			      + 10);
 	(void) sprintf (repository, "%s/%s", CVSroot_directory, argv[0]);
 	repos_len = strlen (CVSroot_directory);
     }
     else
     {
+	repository = xmalloc (strlen (argv[0]) + 5);
 	(void) strcpy (repository, argv[0]);
 	repos_len = 0;
     }
@@ -166,7 +176,7 @@ import (argc, argv)
 	    error (1, 0, "%s is not a numeric branch", vbranch);
     if (numdots (vbranch) != 2)
 	error (1, 0, "Only branches with two dots are supported: %s", vbranch);
-    (void) strcpy (vhead, vbranch);
+    vhead = xstrdup (vbranch);
     cp = strrchr (vhead, '.');
     *cp = '\0';
 
@@ -184,11 +194,12 @@ import (argc, argv)
 	do_editor ((char *) NULL, &message, repository,
 		   (List *) NULL);
     }
-
+    do_verify (message, repository);
     msglen = message == NULL ? 0 : strlen (message);
     if (msglen == 0 || message[msglen - 1] != '\n')
     {
 	char *nm = xmalloc (msglen + 2);
+	*nm = '\0';
 	if (message != NULL)
 	{
 	    (void) strcpy (nm, message);
@@ -249,7 +260,11 @@ import (argc, argv)
     tmpfile = cvs_temp_name ();
     if ((logfp = CVS_FOPEN (tmpfile, "w+")) == NULL)
 	error (1, errno, "cannot create temporary file `%s'", tmpfile);
-    (void) CVS_UNLINK (tmpfile);		/* to be sure it goes away */
+    /* On systems where we can unlink an open file, do so, so it will go
+       away no matter how we exit.  FIXME-maybe: Should be checking for
+       errors but I'm not sure which error(s) we get if we are on a system
+       where one can't unlink open files.  */
+    (void) CVS_UNLINK (tmpfile);
     (void) fprintf (logfp, "\nVendor Tag:\t%s\n", argv[1]);
     (void) fprintf (logfp, "Release Tags:\t");
     for (i = 2; i < argc; i++)
@@ -304,19 +319,25 @@ import (argc, argv)
     li = (struct logfile_info *) xmalloc (sizeof (struct logfile_info));
     li->type = T_TITLE;
     li->tag = xstrdup (vbranch);
+    li->rev_old = li->rev_new = NULL;
     p->data = (char *) li;
     (void) addnode (ulist, p);
     Update_Logfile (repository, message, logfp, ulist);
     dellist (&ulist);
-    (void) fclose (logfp);
+    if (fclose (logfp) < 0)
+	error (0, errno, "error closing %s", tmpfile);
 
     /* Make sure the temporary file goes away, even on systems that don't let
        you delete a file that's in use.  */
-    CVS_UNLINK (tmpfile);
+    if (CVS_UNLINK (tmpfile) < 0 && !existence_error (errno))
+	error (0, errno, "cannot remove %s", tmpfile);
     free (tmpfile);
 
     if (message)
 	free (message);
+    free (repository);
+    free (vbranch);
+    free (vhead);
 
     return (err);
 }
@@ -396,8 +417,10 @@ import_descend (message, vtag, targc, targv)
 #ifdef CLIENT_SUPPORT
 		if (client_active)
 		    err += client_process_import_file (message, dp->d_name,
-							   vtag, targc, targv,
-							   repository);
+                                                       vtag, targc, targv,
+                                                       repository,
+                                                       keyword_opt != NULL &&
+                                                       keyword_opt[0] == 'b');
 		else
 #endif
 		    err += process_import_file (message, dp->d_name,
@@ -434,28 +457,76 @@ process_import_file (message, vfile, vtag, targc, targv)
     int targc;
     char *targv[];
 {
-    char attic_name[PATH_MAX];
-    char rcs[PATH_MAX];
+    char *rcs;
     int inattic = 0;
 
+    rcs = xmalloc (strlen (repository) + strlen (vfile) + sizeof (RCSEXT)
+		   + 5);
     (void) sprintf (rcs, "%s/%s%s", repository, vfile, RCSEXT);
     if (!isfile (rcs))
     {
+	char *attic_name;
+
+	attic_name = xmalloc (strlen (repository) + strlen (vfile) +
+			      sizeof (CVSATTIC) + sizeof (RCSEXT) + 10);
 	(void) sprintf (attic_name, "%s/%s/%s%s", repository, CVSATTIC,
 			vfile, RCSEXT);
 	if (!isfile (attic_name))
 	{
+	    int retval;
+	    char *free_opt = NULL;
+	    char *our_opt = keyword_opt;
 
+	    free (attic_name);
 	    /*
 	     * A new import source file; it doesn't exist as a ,v within the
 	     * repository nor in the Attic -- create it anew.
 	     */
 	    add_log ('N', vfile);
-	    return (add_rcs_file (message, rcs, vfile, vtag, targc, targv));
+
+#ifdef SERVER_SUPPORT
+	    /* The most reliable information on whether the file is binary
+	       is what the client told us.  That is because if the client had
+	       the wrong idea about binaryness, it corrupted the file, so
+	       we might as well believe the client.  */
+	    if (server_active)
+	    {
+		Node *node;
+		List *entries;
+
+		/* Reading all the entries for each file is fairly silly, and
+		   probably slow.  But I am too lazy at the moment to do
+		   anything else.  */
+		entries = Entries_Open (0, NULL);
+		node = findnode_fn (entries, vfile);
+		if (node != NULL)
+		{
+		    Entnode *entdata = (Entnode *) node->data;
+		    if (entdata->type == ENT_FILE)
+		    {
+			assert (entdata->options[0] == '-'
+				&& entdata->options[1] == 'k');
+			our_opt = xstrdup (entdata->options + 2);
+			free_opt = our_opt;
+		    }
+		}
+		Entries_Close (entries);
+	    }
+#endif
+
+	    retval = add_rcs_file (message, rcs, vfile, vhead, our_opt,
+				   vbranch, vtag, targc, targv,
+				   NULL, 0, logfp);
+	    if (free_opt != NULL)
+		free (free_opt);
+	    free (rcs);
+	    return retval;
 	}
+	free (attic_name);
 	inattic = 1;
     }
 
+    free (rcs);
     /*
      * an rcs file exists. have to do things the official, slow, way.
      */
@@ -477,7 +548,6 @@ update_rcs_file (message, vfile, vtag, targc, targv, inattic)
 {
     Vers_TS *vers;
     int letter;
-    int ierrno;
     char *tocvsPath;
     struct file_info finfo;
 
@@ -494,11 +564,7 @@ update_rcs_file (message, vfile, vtag, targc, targv, inattic)
     if (vers->vn_rcs != NULL
 	&& !RCS_isdead(vers->srcfile, vers->vn_rcs))
     {
-	char xtmpfile[PATH_MAX];
 	int different;
-	int retcode = 0;
-
-	(void) sprintf (xtmpfile, "%s/cvs-imp%ld", Tmpdir, (long) getpid());
 
 	/*
 	 * The rcs file does have a revision on the vendor branch. Compare
@@ -509,34 +575,14 @@ update_rcs_file (message, vfile, vtag, targc, targv, inattic)
 	 * This is to try to cut down the number of "C" conflict messages for
 	 * locally modified import source files.
 	 */
-	retcode = RCS_checkout (vers->srcfile, (char *) NULL, vers->vn_rcs,
-				(char *) NULL,
-#ifdef HAVE_RCS5
-				"-ko",
-#else
-				NULL,
-#endif
-				xtmpfile);
-	if (retcode != 0)
-	{
-	    ierrno = errno;
-	    fperror (logfp, 0, retcode == -1 ? ierrno : 0,
-		     "ERROR: cannot co revision %s of file %s", vers->vn_rcs,
-		     vers->srcfile->path);
-	    error (0, retcode == -1 ? ierrno : 0,
-		   "ERROR: cannot co revision %s of file %s", vers->vn_rcs,
-		   vers->srcfile->path);
-	    (void) unlink_file (xtmpfile);
-	    return (1);
-	}
-
 	tocvsPath = wrap_tocvs_process_file (vfile);
-	different = xcmp (xtmpfile, vfile);
+	/* FIXME: Why don't we pass tocvsPath to RCS_cmp_file if it is
+           not NULL?  */
+	different = RCS_cmp_file (vers->srcfile, vers->vn_rcs, "-ko", vfile);
 	if (tocvsPath)
 	    if (unlink_file_dir (tocvsPath) < 0)
 		error (0, errno, "cannot remove %s", tocvsPath);
 
-	(void) unlink_file (xtmpfile);
 	if (!different)
 	{
 	    int retval = 0;
@@ -599,12 +645,14 @@ add_rev (message, rcs, vfile, vers)
 	/* Before RCS_lock existed, we were directing stdout, as well as
 	   stderr, from the RCS command, to DEVNULL.  I wouldn't guess that
 	   was necessary, but I don't know for sure.  */
-        if (RCS_lock (rcs, vbranch, 1) != 0)
-	{
-	    error (0, errno, "fork failed");
-	    return (1);
-	}
+	/* Earlier versions of this function printed a `fork failed' error
+	   when RCS_lock returned an error code.  That's not appropriate
+	   now that RCS_lock is librarified, but should the error text be
+	   preserved? */
+	if (RCS_lock (rcs, vbranch, 1) != 0)
+	    return 1;
 	locked = 1;
+	RCS_rewrite (rcs, NULL, NULL);
     }
     tocvsPath = wrap_tocvs_process_file (vfile);
     if (tocvsPath == NULL)
@@ -629,7 +677,7 @@ add_rev (message, rcs, vfile, vers)
 	}
     }
 
-    status = RCS_checkin (rcs->path, tocvsPath == NULL ? vfile : tocvsPath,
+    status = RCS_checkin (rcs, tocvsPath == NULL ? vfile : tocvsPath,
 			  message, vbranch,
 			  (RCS_FLAGS_QUIET
 			   | (use_file_modtime ? RCS_FLAGS_MODTIME : 0)));
@@ -653,6 +701,7 @@ add_rev (message, rcs, vfile, vers)
 	if (locked)
 	{
 	    (void) RCS_unlock(rcs, vbranch, 0);
+	    RCS_rewrite (rcs, NULL, NULL);
 	}
 	return (1);
     }
@@ -690,6 +739,7 @@ add_tags (rcs, vfile, vtag, targc, targv)
 	       "ERROR: Failed to set tag %s in %s", vtag, rcs->path);
 	return (1);
     }
+    RCS_rewrite (rcs, NULL, NULL);
 
     memset (&finfo, 0, sizeof finfo);
     finfo.file = vfile;
@@ -702,7 +752,9 @@ add_tags (rcs, vfile, vtag, targc, targv)
     vers = Version_TS (&finfo, NULL, vtag, NULL, 1, 0);
     for (i = 0; i < targc; i++)
     {
-	if ((retcode = RCS_settag (rcs, targv[i], vers->vn_rcs)) != 0)
+	if ((retcode = RCS_settag (rcs, targv[i], vers->vn_rcs)) == 0)
+	    RCS_rewrite (rcs, NULL, NULL);
+	else
 	{
 	    ierrno = errno;
 	    fperror (logfp, 0, retcode == -1 ? ierrno : 0,
@@ -832,9 +884,11 @@ get_comment (user)
     char *user;
 {
     char *cp, *suffix;
-    char suffix_path[PATH_MAX];
+    char *suffix_path;
     int i;
+    char *retval;
 
+    suffix_path = xmalloc (strlen (user) + 5);
     cp = strrchr (user, '.');
     if (cp != NULL)
     {
@@ -854,41 +908,93 @@ get_comment (user)
 	suffix = "";			/* will use the default */
     for (i = 0;; i++)
     {
-	if (comtable[i].suffix == NULL)	/* default */
-	    return (comtable[i].comlead);
+	if (comtable[i].suffix == NULL)
+	{
+	    /* Default.  Note we'll always hit this case before we
+	       ever return NULL.  */
+	    retval = comtable[i].comlead;
+	    break;
+	}
 	if (strcmp (suffix, comtable[i].suffix) == 0)
-	    return (comtable[i].comlead);
+	{
+	    retval = comtable[i].comlead;
+	    break;
+	}
     }
+    free (suffix_path);
+    return retval;
 }
 
-static int
-add_rcs_file (message, rcs, user, vtag, targc, targv)
+/* Create a new RCS file from scratch.
+
+   This probably should be moved to rcs.c now that it is called from
+   places outside import.c.
+
+   Return value is 0 for success, or nonzero for failure (in which
+   case an error message will have already been printed).  */
+int
+add_rcs_file (message, rcs, user, add_vhead, key_opt,
+	      add_vbranch, vtag, targc, targv,
+	      desctext, desclen, add_logfp)
+    /* Log message for the addition.  Not used if add_vhead == NULL.  */
     char *message;
+    /* Filename of the RCS file to create.  */
     char *rcs;
+    /* Filename of the file to serve as the contents of the initial
+       revision.  Even if add_vhead is NULL, we use this to determine
+       the modes to give the new RCS file.  */
     char *user;
+
+    /* Revision number of head that we are adding.  Normally 1.1 but
+       could be another revision as long as ADD_VBRANCH is a branch
+       from it.  If NULL, then just add an empty file without any
+       revisions (similar to the one created by "rcs -i").  */
+    char *add_vhead;
+
+    /* Keyword expansion mode, e.g., "b" for binary.  NULL means the
+       default behavior.  */
+    char *key_opt;
+
+    /* Vendor branch to import to, or NULL if none.  If non-NULL, then
+       vtag should also be non-NULL.  */
+    char *add_vbranch;
     char *vtag;
     int targc;
     char *targv[];
+
+    /* If non-NULL, description for the file.  If NULL, the description
+       will be empty.  */
+    char *desctext;
+    size_t desclen;
+
+    /* Write errors to here as well as via error (), or NULL if we should
+       use only error ().  */
+    FILE *add_logfp;
 {
     FILE *fprcs, *fpuser;
     struct stat sb;
     struct tm *ftm;
     time_t now;
-    char altdate1[50];
-#ifndef HAVE_RCS5
-    char altdate2[50];
-#endif
+    char altdate1[MAXDATELEN];
     char *author;
     int i, ierrno, err = 0;
     mode_t mode;
     char *tocvsPath;
     char *userfile;
-    char *local_opt = keyword_opt;
+    char *local_opt = key_opt;
     char *free_opt = NULL;
+    mode_t file_type;
 
     if (noexec)
 	return (0);
 
+    /* Note that as the code stands now, the -k option overrides any
+       settings in wrappers (whether CVSROOT/cvswrappers, -W, or
+       whatever).  Some have suggested this should be the other way
+       around.  As far as I know the documentation doesn't say one way
+       or the other.  Before making a change of this sort, should think
+       about what is best, document it (in cvs.texinfo and NEWS), &c.  */
+
     if (local_opt == NULL)
     {
 	if (wrap_name_has (user, WRAP_RCSOPTION))
@@ -899,14 +1005,45 @@ add_rcs_file (message, rcs, user, vtag, targc, targv)
 
     tocvsPath = wrap_tocvs_process_file (user);
     userfile = (tocvsPath == NULL ? user : tocvsPath);
-    fpuser = CVS_FOPEN (userfile, "r");
-    if (fpuser == NULL)
+
+    /* Opening in text mode is probably never the right thing for the
+       server (because the protocol encodes text files in a fashion
+       which does not depend on what the client or server OS is, as
+       documented in cvsclient.texi), but as long as the server just
+       runs on unix it is a moot point.  */
+
+    /* If PreservePermissions is set, then make sure that the file
+       is a plain file before trying to open it.  Longstanding (although
+       often unpopular) CVS behavior has been to follow symlinks, so we
+       maintain that behavior if PreservePermissions is not on.
+
+       NOTE: this error message used to be `cannot fstat', but is now
+       `cannot lstat'.  I don't see a way around this, since we must
+       stat the file before opening it. -twp */
+
+    if (CVS_LSTAT (userfile, &sb) < 0)
+	error (1, errno, "cannot lstat %s", user);
+    file_type = sb.st_mode & S_IFMT;
+
+    fpuser = NULL;
+    if (!preserve_perms || file_type == S_IFREG)
     {
-	/* not fatal, continue import */
-	fperror (logfp, 0, errno, "ERROR: cannot read file %s", userfile);
-	error (0, errno, "ERROR: cannot read file %s", userfile);
-	goto read_error;
+	fpuser = CVS_FOPEN (userfile,
+			    ((local_opt != NULL && strcmp (local_opt, "b") == 0)
+			     ? "rb"
+			     : "r")
+	    );
+	if (fpuser == NULL)
+	{
+	    /* not fatal, continue import */
+	    if (add_logfp != NULL)
+		fperror (add_logfp, 0, errno,
+			 "ERROR: cannot read file %s", userfile);
+	    error (0, errno, "ERROR: cannot read file %s", userfile);
+	    goto read_error;
+	}
     }
+
     fprcs = CVS_FOPEN (rcs, "w+b");
     if (fprcs == NULL)
     {
@@ -917,20 +1054,45 @@ add_rcs_file (message, rcs, user, vtag, targc, targv)
     /*
      * putadmin()
      */
-    if (fprintf (fprcs, "head     %s;\012", vhead) < 0 ||
-	fprintf (fprcs, "branch   %s;\012", vbranch) < 0 ||
-	fprintf (fprcs, "access   ;\012") < 0 ||
+    if (add_vhead != NULL)
+    {
+	if (fprintf (fprcs, "head     %s;\012", add_vhead) < 0)
+	    goto write_error;
+    }
+    else
+    {
+	if (fprintf (fprcs, "head     ;\012") < 0)
+	    goto write_error;
+    }
+
+    if (add_vbranch != NULL)
+    {
+	if (fprintf (fprcs, "branch   %s;\012", add_vbranch) < 0)
+	    goto write_error;
+    }
+    if (fprintf (fprcs, "access   ;\012") < 0 ||
 	fprintf (fprcs, "symbols  ") < 0)
     {
 	goto write_error;
     }
 
-    for (i = targc - 1; i >= 0; i--)	/* RCS writes the symbols backwards */
-	if (fprintf (fprcs, "%s:%s.1 ", targv[i], vbranch) < 0)
+    for (i = targc - 1; i >= 0; i--)
+    {
+	/* RCS writes the symbols backwards */
+	assert (add_vbranch != NULL);
+	if (fprintf (fprcs, "%s:%s.1 ", targv[i], add_vbranch) < 0)
 	    goto write_error;
+    }
 
-    if (fprintf (fprcs, "%s:%s;\012", vtag, vbranch) < 0 ||
-	fprintf (fprcs, "locks    ; strict;\012") < 0 ||
+    if (add_vbranch != NULL)
+    {
+	if (fprintf (fprcs, "%s:%s", vtag, add_vbranch) < 0)
+	    goto write_error;
+    }
+    if (fprintf (fprcs, ";\012") < 0)
+	goto write_error;
+
+    if (fprintf (fprcs, "locks    ; strict;\012") < 0 ||
 	/* XXX - make sure @@ processing works in the RCS file */
 	fprintf (fprcs, "comment  @%s@;\012", get_comment (user)) < 0)
     {
@@ -948,99 +1110,225 @@ add_rcs_file (message, rcs, user, vtag, targc, targv)
     if (fprintf (fprcs, "\012") < 0)
       goto write_error;
 
-    /*
-     * puttree()
-     */
-    if (fstat (fileno (fpuser), &sb) < 0)
-	error (1, errno, "cannot fstat %s", user);
-    if (use_file_modtime)
-	now = sb.st_mtime;
-    else
-	(void) time (&now);
-#ifdef HAVE_RCS5
-    ftm = gmtime (&now);
-#else
-    ftm = localtime (&now);
-#endif
-    (void) sprintf (altdate1, DATEFORM,
-		    ftm->tm_year + (ftm->tm_year < 100 ? 0 : 1900),
-		    ftm->tm_mon + 1, ftm->tm_mday, ftm->tm_hour,
-		    ftm->tm_min, ftm->tm_sec);
-#ifdef HAVE_RCS5
-#define	altdate2 altdate1
-#else
-    /*
-     * If you don't have RCS V5 or later, you need to lie about the ci
-     * time, since RCS V4 and earlier insist that the times differ.
-     */
-    now++;
-    ftm = localtime (&now);
-    (void) sprintf (altdate2, DATEFORM,
-		    ftm->tm_year + (ftm->tm_year < 100 ? 0 : 1900),
-		    ftm->tm_mon + 1, ftm->tm_mday, ftm->tm_hour,
-		    ftm->tm_min, ftm->tm_sec);
-#endif
-    author = getcaller ();
-
-    if (fprintf (fprcs, "\012%s\012", vhead) < 0 ||
-	fprintf (fprcs, "date     %s;  author %s;  state Exp;\012",
-		 altdate1, author) < 0 ||
-	fprintf (fprcs, "branches %s.1;\012", vbranch) < 0 ||
-	fprintf (fprcs, "next     ;\012") < 0 ||
-	fprintf (fprcs, "\012%s.1\012", vbranch) < 0 ||
-	fprintf (fprcs, "date     %s;  author %s;  state Exp;\012",
-		 altdate2, author) < 0 ||
-	fprintf (fprcs, "branches ;\012") < 0 ||
-	fprintf (fprcs, "next     ;\012\012") < 0 ||
-	/*
-	 * putdesc()
-	 */
-	fprintf (fprcs, "\012desc\012") < 0 ||
-	fprintf (fprcs, "@@\012\012\012") < 0 ||
-	/*
-	 * putdelta()
-	 */
-	fprintf (fprcs, "\012%s\012", vhead) < 0 ||
-	fprintf (fprcs, "log\012") < 0 ||
-	fprintf (fprcs, "@Initial revision\012@\012") < 0 ||
-	fprintf (fprcs, "text\012@") < 0)
+    /* Write the revision(s), with the date and author and so on
+       (that is "delta" rather than "deltatext" from rcsfile(5)).  */
+    if (add_vhead != NULL)
     {
+	if (use_file_modtime)
+	    now = sb.st_mtime;
+	else
+	    (void) time (&now);
+	ftm = gmtime (&now);
+	(void) sprintf (altdate1, DATEFORM,
+			ftm->tm_year + (ftm->tm_year < 100 ? 0 : 1900),
+			ftm->tm_mon + 1, ftm->tm_mday, ftm->tm_hour,
+			ftm->tm_min, ftm->tm_sec);
+	author = getcaller ();
+
+	if (fprintf (fprcs, "\012%s\012", add_vhead) < 0 ||
+	fprintf (fprcs, "date     %s;  author %s;  state Exp;\012",
+		 altdate1, author) < 0)
 	goto write_error;
-    }
 
-    /* Now copy over the contents of the file, expanding at signs.  */
-    {
-	char buf[8192];
-	unsigned int len;
-
-	while (1)
+	if (fprintf (fprcs, "branches") < 0)
+	    goto write_error;
+	if (add_vbranch != NULL)
 	{
-	    len = fread (buf, 1, sizeof buf, fpuser);
-	    if (len == 0)
+	    if (fprintf (fprcs, " %s.1", add_vbranch) < 0)
+		goto write_error;
+	}
+	if (fprintf (fprcs, ";\012") < 0)
+	    goto write_error;
+
+	if (fprintf (fprcs, "next     ;\012") < 0)
+	    goto write_error;
+
+#ifdef PRESERVE_PERMISSIONS_SUPPORT
+	/* Store initial permissions if necessary. */
+	if (preserve_perms)
+	{
+	    if (file_type == S_IFLNK)
 	    {
-		if (ferror (fpuser))
-		    error (1, errno, "cannot read file %s for copying", user);
-		break;
+		char *link = xreadlink (userfile);
+		if (fprintf (fprcs, "symlink\t@") < 0 ||
+		    expand_at_signs (link, strlen (link), fprcs) < 0 ||
+		    fprintf (fprcs, "@;\012") < 0)
+		    goto write_error;
+		free (link);
 	    }
-	    if (expand_at_signs (buf, len, fprcs) < 0)
+	    else
+	    {
+		if (fprintf (fprcs, "owner\t%u;\012", sb.st_uid) < 0)
+		    goto write_error;
+		if (fprintf (fprcs, "group\t%u;\012", sb.st_gid) < 0)
+		    goto write_error;
+		if (fprintf (fprcs, "permissions\t%o;\012",
+			     sb.st_mode & 07777) < 0)
+		    goto write_error;
+		switch (file_type)
+		{
+		    case S_IFREG: break;
+		    case S_IFCHR:
+		    case S_IFBLK:
+			if (fprintf (fprcs, "special\t%s %lu;\012",
+				     (file_type == S_IFCHR
+				      ? "character"
+				      : "block"),
+				     (unsigned long) sb.st_rdev) < 0)
+			    goto write_error;
+			break;
+		    default:
+			error (0, 0,
+			       "can't import %s: unknown kind of special file",
+			       userfile);
+		}
+	    }
+	}
+#endif
+
+	if (add_vbranch != NULL)
+	{
+	    if (fprintf (fprcs, "\012%s.1\012", add_vbranch) < 0 ||
+		fprintf (fprcs, "date     %s;  author %s;  state Exp;\012",
+			 altdate1, author) < 0 ||
+		fprintf (fprcs, "branches ;\012") < 0 ||
+		fprintf (fprcs, "next     ;\012") < 0)
+		goto write_error;
+
+#ifdef PRESERVE_PERMISSIONS_SUPPORT
+	    /* Store initial permissions if necessary. */
+	    if (preserve_perms)
+	    {
+		if (file_type == S_IFLNK)
+		{
+		    char *link = xreadlink (userfile);
+		    if (fprintf (fprcs, "symlink\t@") < 0 ||
+			expand_at_signs (link, strlen (link), fprcs) < 0 ||
+			fprintf (fprcs, "@;\012") < 0)
+			goto write_error;
+		    free (link);
+		}
+		else
+		{
+		    if (fprintf (fprcs, "owner\t%u;\012", sb.st_uid) < 0 ||
+			fprintf (fprcs, "group\t%u;\012", sb.st_gid) < 0 ||
+			fprintf (fprcs, "permissions\t%o;\012",
+				 sb.st_mode & 07777) < 0)
+			goto write_error;
+	    
+		    switch (file_type)
+		    {
+			case S_IFREG: break;
+			case S_IFCHR:
+			case S_IFBLK:
+			    if (fprintf (fprcs, "special\t%s %lu;\012",
+					 (file_type == S_IFCHR
+					  ? "character"
+					  : "block"),
+					 (unsigned long) sb.st_rdev) < 0)
+				goto write_error;
+			    break;
+			default:
+			    error (0, 0,
+			      "cannot import %s: special file of unknown type",
+			       userfile);
+		    }
+		}
+	    }
+#endif
+
+	    if (fprintf (fprcs, "\012") < 0)
 		goto write_error;
 	}
     }
-    if (fprintf (fprcs, "@\012\012") < 0 ||
-	fprintf (fprcs, "\012%s.1\012", vbranch) < 0 ||
-	fprintf (fprcs, "log\012@") < 0 ||
-	expand_at_signs (message, (off_t) strlen (message), fprcs) < 0 ||
-	fprintf (fprcs, "@\012text\012") < 0 ||
-	fprintf (fprcs, "@@\012") < 0)
-    {
+
+    /* Now write the description (possibly empty).  */
+    if (fprintf (fprcs, "\012desc\012") < 0 ||
+	fprintf (fprcs, "@") < 0)
 	goto write_error;
+    if (desctext != NULL)
+    {
+	/* The use of off_t not size_t for the second argument is very
+	   strange, since we are dealing with something which definitely
+	   fits in memory.  */
+	if (expand_at_signs (desctext, (off_t) desclen, fprcs) < 0)
+	    goto write_error;
     }
+    if (fprintf (fprcs, "@\012\012\012") < 0)
+	goto write_error;
+
+    /* Now write the log messages and contents for the revision(s) (that
+       is, "deltatext" rather than "delta" from rcsfile(5)).  */
+    if (add_vhead != NULL)
+    {
+	if (fprintf (fprcs, "\012%s\012", add_vhead) < 0 ||
+	    fprintf (fprcs, "log\012@") < 0)
+	    goto write_error;
+	if (add_vbranch != NULL)
+	{
+	    /* We are going to put the log message in the revision on the
+	       branch.  So putting it here too seems kind of redundant, I
+	       guess (and that is what CVS has always done, anyway).  */
+	    if (fprintf (fprcs, "Initial revision\012") < 0)
+		goto write_error;
+	}
+	else
+	{
+	    if (expand_at_signs (message, (off_t) strlen (message), fprcs) < 0)
+		goto write_error;
+	}
+	if (fprintf (fprcs, "@\012") < 0 ||
+	    fprintf (fprcs, "text\012@") < 0)
+	{
+	    goto write_error;
+	}
+
+	/* Now copy over the contents of the file, expanding at signs.
+	   If preserve_perms is set, do this only for regular files. */
+	if (!preserve_perms || file_type == S_IFREG)
+	{
+	    char buf[8192];
+	    unsigned int len;
+
+	    while (1)
+	    {
+		len = fread (buf, 1, sizeof buf, fpuser);
+		if (len == 0)
+		{
+		    if (ferror (fpuser))
+			error (1, errno, "cannot read file %s for copying",
+			       user);
+		    break;
+		}
+		if (expand_at_signs (buf, len, fprcs) < 0)
+		    goto write_error;
+	    }
+	}
+	if (fprintf (fprcs, "@\012\012") < 0)
+	    goto write_error;
+	if (add_vbranch != NULL)
+	{
+	    if (fprintf (fprcs, "\012%s.1\012", add_vbranch) < 0 ||
+		fprintf (fprcs, "log\012@") < 0 ||
+		expand_at_signs (message,
+				 (off_t) strlen (message), fprcs) < 0 ||
+		fprintf (fprcs, "@\012text\012") < 0 ||
+		fprintf (fprcs, "@@\012") < 0)
+		goto write_error;
+	}
+    }
+
     if (fclose (fprcs) == EOF)
     {
 	ierrno = errno;
 	goto write_error_noclose;
     }
-    (void) fclose (fpuser);
+    /* Close fpuser only if we opened it to begin with. */
+    if (fpuser != NULL)
+    {
+	if (fclose (fpuser) < 0)
+	    error (0, errno, "cannot close %s", user);
+    }
 
     /*
      * Fix the modes on the RCS files.  The user modes of the original
@@ -1056,8 +1344,9 @@ add_rcs_file (message, rcs, user, vtag, targc, targv)
     if (chmod (rcs, mode) < 0)
     {
 	ierrno = errno;
-	fperror (logfp, 0, ierrno,
-		 "WARNING: cannot change mode of file %s", rcs);
+	if (add_logfp != NULL)
+	    fperror (add_logfp, 0, ierrno,
+		     "WARNING: cannot change mode of file %s", rcs);
 	error (0, ierrno, "WARNING: cannot change mode of file %s", rcs);
 	err++;
     }
@@ -1070,15 +1359,20 @@ add_rcs_file (message, rcs, user, vtag, targc, targv)
 
 write_error:
     ierrno = errno;
-    (void) fclose (fprcs);
+    if (fclose (fprcs) < 0)
+	error (0, errno, "cannot close %s", rcs);
 write_error_noclose:
-    (void) fclose (fpuser);
-    fperror (logfp, 0, ierrno, "ERROR: cannot write file %s", rcs);
+    if (fclose (fpuser) < 0)
+	error (0, errno, "cannot close %s", user);
+    if (add_logfp != NULL)
+	fperror (add_logfp, 0, ierrno, "ERROR: cannot write file %s", rcs);
     error (0, ierrno, "ERROR: cannot write file %s", rcs);
     if (ierrno == ENOSPC)
     {
-	(void) CVS_UNLINK (rcs);
-	fperror (logfp, 0, 0, "ERROR: out of space - aborting");
+	if (CVS_UNLINK (rcs) < 0)
+	    error (0, errno, "cannot remove %s", rcs);
+	if (add_logfp != NULL)
+	    fperror (add_logfp, 0, 0, "ERROR: out of space - aborting");
 	error (1, 0, "ERROR: out of space - aborting");
     }
 read_error:
@@ -1097,26 +1391,33 @@ read_error:
  * signs.  If an error occurs, return a negative value and set errno
  * to indicate the error.  If not, return a nonnegative value.
  */
-static int
+int
 expand_at_signs (buf, size, fp)
     char *buf;
     off_t size;
     FILE *fp;
 {
-    char *cp, *end;
+    register char *cp, *next;
 
-    errno = 0;
-    for (cp = buf, end = buf + size; cp < end; cp++)
+    cp = buf;
+    while ((next = memchr (cp, '@', size)) != NULL)
     {
-	if (*cp == '@')
-	{
-	    if (putc ('@', fp) == EOF && errno != 0)
-		return EOF;
-	}
-	if (putc (*cp, fp) == EOF && errno != 0)
-	    return (EOF);
+	int len;
+
+	++next;
+	len = next - cp;
+	if (fwrite (cp, 1, len, fp) != len)
+	    return EOF;
+	if (putc ('@', fp) == EOF)
+	    return EOF;
+	cp = next;
+	size -= len;
     }
-    return (1);
+
+    if (fwrite (cp, 1, size, fp) != size)
+	return EOF;
+
+    return 1;
 }
 
 /*
@@ -1174,6 +1475,7 @@ import_descend_dir (message, dir, vtag, targc, targv)
     struct saved_cwd cwd;
     char *cp;
     int ierrno, err;
+    char *rcs = NULL;
 
     if (islink (dir))
 	return (0);
@@ -1182,13 +1484,24 @@ import_descend_dir (message, dir, vtag, targc, targv)
 	fperror (logfp, 0, 0, "ERROR: cannot get working directory");
 	return (1);
     }
+
+    /* Concatenate DIR to the end of REPOSITORY.  */
     if (repository[0] == '\0')
-	(void) strcpy (repository, dir);
+    {
+	char *new = xstrdup (dir);
+	free (repository);
+	repository = new;
+    }
     else
     {
-	(void) strcat (repository, "/");
-	(void) strcat (repository, dir);
+	char *new = xmalloc (strlen (repository) + strlen (dir) + 10);
+	strcpy (new, repository);
+	(void) strcat (new, "/");
+	(void) strcat (new, dir);
+	free (repository);
+	repository = new;
     }
+
 #ifdef CLIENT_SUPPORT
     if (!quiet && !client_active)
 #else
@@ -1210,12 +1523,12 @@ import_descend_dir (message, dir, vtag, targc, targv)
     if (!isdir (repository))
 #endif
     {
-	char rcs[PATH_MAX];
-
+	rcs = xmalloc (strlen (repository) + sizeof (RCSEXT) + 5);
 	(void) sprintf (rcs, "%s%s", repository, RCSEXT);
 	if (isfile (repository) || isfile(rcs))
 	{
-	    fperror (logfp, 0, 0, "ERROR: %s is a file, should be a directory!",
+	    fperror (logfp, 0, 0,
+		     "ERROR: %s is a file, should be a directory!",
 		     repository);
 	    error (0, 0, "ERROR: %s is a file, should be a directory!",
 		   repository);
@@ -1235,12 +1548,14 @@ import_descend_dir (message, dir, vtag, targc, targv)
     }
     err = import_descend (message, vtag, targc, targv);
   out:
+    if (rcs != NULL)
+	free (rcs);
     if ((cp = strrchr (repository, '/')) != NULL)
 	*cp = '\0';
     else
 	repository[0] = '\0';
     if (restore_cwd (&cwd, NULL))
-      exit (EXIT_FAILURE);
+	error_exit ();
     free_cwd (&cwd);
     return (err);
 }
