@@ -704,7 +704,6 @@ lock_obtained (repos)
 static int lock_filesdoneproc PROTO ((void *callerdat, int err,
 				      char *repository, char *update_dir,
 				      List *entries));
-static int fsortcmp PROTO((const Node * p, const Node * q));
 
 /*
  * Create a list of repositories to lock
@@ -733,17 +732,6 @@ lock_filesdoneproc (callerdat, err, repository, update_dir, entries)
     return (err);
 }
 
-/*
- * compare two lock list nodes (for sort)
- */
-static int
-fsortcmp (p, q)
-    const Node *p;
-    const Node *q;
-{
-    return (strcmp (p->key, q->key));
-}
-
 void
 lock_tree_for_write (argc, argv, local, aflag)
     int argc;
