@@ -3,7 +3,7 @@
  *    Copyright (c) 1989-1992, Brian Berliner
  *
  *    You may distribute under the terms of the GNU General Public License
- *    as specified in the README file that comes with the CVS 1.4 kit.
+ *    as specified in the README file that comes with the CVS source distribution.
  *
  * This is the main C driver for the CVS system.
  *
@@ -24,23 +24,23 @@ char *program_name;
 char *program_path;
 char *command_name;
 
-/*
- * Since some systems don't define this...
- */
+/* I'd dynamically allocate this, but it seems like gethostname
+   requires a fixed size array.  If I'm remembering the RFCs right,
+   256 should be enough.  */
 #ifndef MAXHOSTNAMELEN
 #define MAXHOSTNAMELEN  256
 #endif
 
 char hostname[MAXHOSTNAMELEN];
 
-int use_editor = TRUE;
-int use_cvsrc = TRUE;
+int use_editor = 1;
+int use_cvsrc = 1;
 int cvswrite = !CVSREAD_DFLT;
-int really_quiet = FALSE;
-int quiet = FALSE;
-int trace = FALSE;
-int noexec = FALSE;
-int logoff = FALSE;
+int really_quiet = 0;
+int quiet = 0;
+int trace = 0;
+int noexec = 0;
+int logoff = 0;
 mode_t cvsumask = UMASK_DFLT;
 
 char *CurDir;
@@ -48,30 +48,8 @@ char *CurDir;
 /*
  * Defaults, for the environment variables that are not set
  */
-char *Rcsbin = RCSBIN_DFLT;
 char *Tmpdir = TMPDIR_DFLT;
 char *Editor = EDITOR_DFLT;
-/*
- * The path found in CVS/Root must match $CVSROOT and/or 'cvs -d root'
- */
-int add PROTO((int argc, char **argv));
-int admin PROTO((int argc, char **argv));
-int checkout PROTO((int argc, char **argv));
-int commit PROTO((int argc, char **argv));
-int diff PROTO((int argc, char **argv));
-int history PROTO((int argc, char **argv));
-int import PROTO((int argc, char **argv));
-int cvslog PROTO((int argc, char **argv));
-#ifdef AUTH_CLIENT_SUPPORT
-int login PROTO((int argc, char **argv));
-#endif /* AUTH_CLIENT_SUPPORT */
-int patch PROTO((int argc, char **argv));
-int release PROTO((int argc, char **argv));
-int cvsremove PROTO((int argc, char **argv));
-int rtag PROTO((int argc, char **argv));
-int status PROTO((int argc, char **argv));
-int tag PROTO((int argc, char **argv));
-int update PROTO((int argc, char **argv));
 
 static const struct cmd
 {
@@ -99,7 +77,7 @@ static const struct cmd
 
     char *nick1;
     char *nick2;
-
+    
     int (*func) ();		/* Function takes (argc, argv) arguments. */
 } cmds[] =
 
@@ -122,6 +100,7 @@ static const struct cmd
     { "log",      "lo",       "rlog",      cvslog },
 #ifdef AUTH_CLIENT_SUPPORT
     { "login",    "logon",    "lgn",       login },
+    { "logout",   NULL,       NULL,        logout },
 #ifdef SERVER_SUPPORT
     { "pserver",  NULL,       NULL,        server }, /* placeholder */
 #endif
@@ -131,7 +110,7 @@ static const struct cmd
     { "remove",   "rm",       "delete",    cvsremove },
     { "status",   "st",       "stat",      status },
     { "rtag",     "rt",       "rfreeze",   rtag },
-    { "tag",      "ta",       "freeze",    tag },
+    { "tag",      "ta",       "freeze",    cvstag },
     { "unedit",   NULL,	      NULL,	   unedit },
     { "update",   "up",       "upd",       update },
     { "watch",    NULL,	      NULL,	   watch },
@@ -144,32 +123,47 @@ static const struct cmd
 
 static const char *const usg[] =
 {
-    "Usage: %s [cvs-options] command [command-options] [files...]\n",
-    "    Where 'cvs-options' are:\n",
-    "        -H           Displays Usage information for command\n",
-    "        -Q           Cause CVS to be really quiet.\n",
-    "        -q           Cause CVS to be somewhat quiet.\n",
-    "        -r           Make checked-out files read-only\n",
-    "        -w           Make checked-out files read-write (default)\n",
-    "        -l           Turn History logging off\n",
-    "        -n           Do not execute anything that will change the disk\n",
-    "        -t           Show trace of program execution -- Try with -n\n",
-    "        -v           CVS version and copyright\n",
-    "        -b bindir    Find RCS programs in 'bindir'\n",
-    "        -T tmpdir    Use 'tmpdir' for temporary files\n",
-    "        -e editor    Use 'editor' for editing log information\n",
-    "        -d CVS_root  Overrides $CVSROOT as the root of the CVS tree\n",
-    "        -f           Do not use the ~/.cvsrc file\n",
-#ifdef CLIENT_SUPPORT
-    "        -z #         Use compression level '#' for net traffic.\n",
-#ifdef ENCRYPTION
-    "        -x           Encrypt all net traffic.\n",
-#endif
-#endif
-    "        -s VAR=VAL   Set CVS user variable.\n",
+    /* CVS usage messages never have followed the GNU convention of
+       putting metavariables in uppercase.  I don't know whether that
+       is a good convention or not, but if it changes it would have to
+       change in all the usage messages.  For now, they consistently
+       use lowercase, as far as I know.  Puncutation is pretty funky,
+       though.  Sometimes they use none, as here.  Sometimes they use
+       single quotes (not the TeX-ish `' stuff), as in --help-options.
+       Sometimes they use double quotes, as in cvs -H add.
+
+       Most (not all) of the usage messages seem to have periods at
+       the end of each line.  I haven't tried to duplicate this style
+       in --help as it is a rather different format from the rest.  */
+
+    "Usage: %s [cvs-options] command [command-options-and-arguments]\n",
+    "  where cvs-options are -q, -n, etc.\n",
+    "    (specify --help-options for a list of options)\n",
+    "  where command is add, admin, etc.\n",
+    "    (specify --help-commands for a list of commands\n",
+    "     or --help-synonyms for a list of command synonyms)\n",
+    "  where command-options-and-arguments depend on the specific command\n",
+    "    (specify -H followed by a command name for command-specific help)\n",
+    "  Specify --help to receive this message\n",
     "\n",
-    "    and where 'command' is: add, admin, etc. (use the --help-commands\n",
-    "    option for a list of commands)\n",
+
+    /* Some people think that a bug-reporting address should go here.  IMHO,
+       the web sites are better because anything else is very likely to go
+       obsolete in the years between a release and when someone might be
+       reading this help.  Besides, we could never adequately discuss
+       bug reporting in a concise enough way to put in a help message.  */
+
+    /* I was going to put this at the top, but usage() wants the %s to
+       be in the first line.  */
+    "The Concurrent Versions System (CVS) is a tool for version control.\n",
+    /* I really don't think I want to try to define "version control"
+       in one line.  I'm not sure one can get more concise than the
+       paragraph in ../cvs.spec without assuming the reader knows what
+       version control means.  */
+
+    "For CVS updates and additional information, see\n",
+    "    Cyclic Software at http://www.cyclic.com/ or\n",
+    "    Pascal Molli's CVS site at http://www.loria.fr/~molli/cvs-index.html\n",
     NULL,
 };
 
@@ -181,7 +175,7 @@ static const char *const cmd_usage[] =
     "        annotate     Show last revision where each line was modified\n",
     "        checkout     Checkout sources for editing\n",
     "        commit       Check files into the repository\n",
-    "        diff         Run diffs between revisions\n",
+    "        diff         Show differences between revisions\n",
     "        edit         Get ready to edit a watched file\n",
     "        editors      See who is editing a watched file\n",
     "        export       Export sources from CVS, similar to checkout\n",
@@ -191,6 +185,7 @@ static const char *const cmd_usage[] =
     "        log          Print out history information for files\n",
 #ifdef AUTH_CLIENT_SUPPORT
     "        login        Prompt for password for authenticating server.\n",
+    "        logout       Removes entry in .cvspass for remote repository.\n",
 #endif /* AUTH_CLIENT_SUPPORT */
     "        rdiff        Create 'patch' format diffs between releases\n",
     "        release      Indicate that a Module is no longer in use\n",
@@ -202,17 +197,47 @@ static const char *const cmd_usage[] =
     "        update       Bring work tree in sync with repository\n",
     "        watch        Set watches\n",
     "        watchers     See who is watching a file\n",
-    "(Use the --help-synonyms option for a list of alternate command names)\n",
+    "(Specify the --help option for a list of other help options)\n",
     NULL,
 };
 
+static const char *const opt_usage[] =
+{
+    "CVS global options (specified before the command name) are:\n",
+    "    -H           Displays usage information for command.\n",
+    "    -Q           Cause CVS to be really quiet.\n",
+    "    -q           Cause CVS to be somewhat quiet.\n",
+    "    -r           Make checked-out files read-only.\n",
+    "    -w           Make checked-out files read-write (default).\n",
+    "    -l           Turn history logging off.\n",
+    "    -n           Do not execute anything that will change the disk.\n",
+    "    -t           Show trace of program execution -- try with -n.\n",
+    "    -v           CVS version and copyright.\n",
+    "    -b bindir    Find RCS programs in 'bindir'.\n",
+    "    -T tmpdir    Use 'tmpdir' for temporary files.\n",
+    "    -e editor    Use 'editor' for editing log information.\n",
+    "    -d CVS_root  Overrides $CVSROOT as the root of the CVS tree.\n",
+    "    -f           Do not use the ~/.cvsrc file.\n",
+#ifdef CLIENT_SUPPORT
+    "    -z #         Use compression level '#' for net traffic.\n",
+#ifdef ENCRYPTION
+    "    -x           Encrypt all net traffic.\n",
+#endif
+    "    -a           Authenticate all net traffic.\n",
+#endif
+    "    -s VAR=VAL   Set CVS user variable.\n",
+    "(Specify the --help option for a list of other help options)\n",
+    NULL
+};
+
 static const char * const*
 cmd_synonyms ()
 {
     char ** synonyms;
     char ** line;
     const struct cmd *c = &cmds[0];
-    int numcmds = 2;		/* two more for title and end */
+    /* Three more for title, "specify --help" line, and NULL.  */
+    int numcmds = 3;
 
     while (c->fullname != NULL)
     {
@@ -227,18 +252,69 @@ cmd_synonyms ()
     {
 	if (c->nick1 || c->nick2)
 	{
-	    *line = xmalloc(100); /* wild guess */
+	    *line = xmalloc (strlen (c->fullname)
+			     + (c->nick1 != NULL ? strlen (c->nick1) : 0)
+			     + (c->nick2 != NULL ? strlen (c->nick2) : 0)
+			     + 40);
 	    sprintf(*line, "        %-12s %s %s\n", c->fullname,
 		    c->nick1 ? c->nick1 : "",
 		    c->nick2 ? c->nick2 : "");
 	    line++;
 	}
     }
+    *line++ = "(Specify the --help option for a list of other help options)\n";
     *line = NULL;
     
     return (const char * const*) synonyms; /* will never be freed */
 }
 
+
+unsigned long int
+lookup_command_attribute (cmd_name)
+     char *cmd_name;
+{
+    unsigned long int ret = 0;
+
+    if (strcmp (cmd_name, "import") != 0)
+    {
+        ret |= CVS_CMD_IGNORE_ADMROOT;
+    }
+
+
+    if ((strcmp (cmd_name, "checkout") != 0) &&
+        (strcmp (cmd_name, "init") != 0) &&
+        (strcmp (cmd_name, "login") != 0) &&
+	(strcmp (cmd_name, "logout") != 0) &&
+        (strcmp (cmd_name, "rdiff") != 0) &&
+        (strcmp (cmd_name, "release") != 0) &&
+        (strcmp (cmd_name, "rtag") != 0))
+    {
+        ret |= CVS_CMD_USES_WORK_DIR;
+    }
+
+
+    /* The following commands do not modify the repository; we
+       conservatively assume that everything else does.  Feel free to
+       add to this list if you are _certain_ something is safe. */
+    if ((strcmp (cmd_name, "checkout") != 0) &&
+        (strcmp (cmd_name, "diff") != 0) &&
+        (strcmp (cmd_name, "update") != 0) &&
+        (strcmp (cmd_name, "history") != 0) &&
+        (strcmp (cmd_name, "editors") != 0) &&
+        (strcmp (cmd_name, "export") != 0) &&
+        (strcmp (cmd_name, "history") != 0) &&
+        (strcmp (cmd_name, "log") != 0) &&
+        (strcmp (cmd_name, "noop") != 0) &&
+        (strcmp (cmd_name, "watchers") != 0) &&
+        (strcmp (cmd_name, "status") != 0))
+    {
+        ret |= CVS_CMD_MODIFIES_REPOSITORY;
+    }
+
+    return ret;
+}
+
+
 static RETSIGTYPE
 main_cleanup (sig)
     int sig;
@@ -286,16 +362,6 @@ main_cleanup (sig)
 #endif /* !DONT_USE_SIGNALS */
 }
 
-static void
-error_cleanup PROTO((void))
-{
-    Lock_Cleanup();
-#ifdef SERVER_SUPPORT
-    if (server_active)
-	server_cleanup (0);
-#endif
-}
-
 int
 main (argc, argv)
     int argc;
@@ -307,31 +373,41 @@ main (argc, argv)
     char *cp, *end;
     const struct cmd *cm;
     int c, err = 0;
-    int rcsbin_update_env, tmpdir_update_env, cvs_update_env;
+    int tmpdir_update_env, cvs_update_env;
+    int free_CVSroot = 0;
+    int free_Editor = 0;
+    int free_Tmpdir = 0;
+
     int help = 0;		/* Has the user asked for help?  This
 				   lets us support the `cvs -H cmd'
 				   convention to give help for cmd. */
     static struct option long_options[] =
-      {
+    {
         {"help", 0, NULL, 'H'},
         {"version", 0, NULL, 'v'},
 	{"help-commands", 0, NULL, 1},
 	{"help-synonyms", 0, NULL, 2},
+	{"help-options", 0, NULL, 4},
+	{"allow-root", required_argument, NULL, 3},
         {0, 0, 0, 0}
-      };
+    };
     /* `getopt_long' stores the option index here, but right now we
         don't use it. */
     int option_index = 0;
     int need_to_create_root = 0;
 
-    error_set_cleanup (error_cleanup);
-
 #ifdef SYSTEM_INITIALIZE
     /* Hook for OS-specific behavior, for example socket subsystems on
        NT and OS2 or dealing with windows and arguments on Mac.  */
     SYSTEM_INITIALIZE (&argc, &argv);
 #endif
 
+#ifdef HAVE_TZSET
+    /* On systems that have tzset (which is almost all the ones I know
+       of), it's a good idea to call it.  */
+    tzset ();
+#endif
+
     /*
      * Just save the last component of the path for error messages
      */
@@ -349,12 +425,6 @@ main (argc, argv)
      * they can be overridden by command line arguments
      */
     cvs_update_env = 0;
-    rcsbin_update_env = *Rcsbin;	/* RCSBIN_DFLT must be set */
-    if ((cp = getenv (RCSBIN_ENV)) != NULL)
-    {
-	Rcsbin = cp;
-	rcsbin_update_env = 0;		/* it's already there */
-    }
     tmpdir_update_env = *Tmpdir;	/* TMPDIR_DFLT must be set */
     if ((cp = getenv (TMPDIR_ENV)) != NULL)
     {
@@ -373,15 +443,11 @@ main (argc, argv)
 	cvs_update_env = 0;		/* it's already there */
     }
     if (getenv (CVSREAD_ENV) != NULL)
-	cvswrite = FALSE;
-
-    /* This has the effect of setting getopt's ordering to REQUIRE_ORDER,
-       which is what we need to distinguish between global options and
-       command options.  FIXME: It would appear to be possible to do this
-       much less kludgily by passing "+" as the first character to the
-       option string we pass to getopt_long.  */
-    optind = 1;
+	cvswrite = 0;
 
+    /* Set this to 0 to force getopt initialization.  getopt() sets
+       this to 1 internally.  */
+    optind = 0;
 
     /* We have to parse the options twice because else there is no
        chance to avoid reading the global options from ".cvsrc".  Set
@@ -390,28 +456,28 @@ main (argc, argv)
     opterr = 0;
 
     while ((c = getopt_long
-            (argc, argv, "f", NULL, NULL))
+            (argc, argv, "+f", NULL, NULL))
            != EOF)
-      {
+    {
 	if (c == 'f')
-	    use_cvsrc = FALSE;
-      }
-    
+	    use_cvsrc = 0;
+    }
+
     /*
      * Scan cvsrc file for global options.
      */
     if (use_cvsrc)
 	read_cvsrc (&argc, &argv, "cvs");
 
-    optind = 1;
+    optind = 0;
     opterr = 1;
 
     while ((c = getopt_long
-            (argc, argv, "Qqrwtnlvb:T:e:d:Hfz:s:x", long_options, &option_index))
+            (argc, argv, "+Qqrwtnlvb:T:e:d:Hfz:s:xa", long_options, &option_index))
            != EOF)
-      {
+    {
 	switch (c)
-          {
+	{
             case 1:
 	        /* --help-commands */
                 usage (cmd_usage);
@@ -420,59 +486,79 @@ main (argc, argv)
 	        /* --help-synonyms */
                 usage (cmd_synonyms());
                 break;
+	    case 4:
+		/* --help-options */
+		usage (opt_usage);
+		break;
+	    case 3:
+		/* --allow-root */
+		root_allow_add (optarg);
+		break;
 	    case 'Q':
-		really_quiet = TRUE;
+		really_quiet = 1;
 		/* FALL THROUGH */
 	    case 'q':
-		quiet = TRUE;
+		quiet = 1;
 		break;
 	    case 'r':
-		cvswrite = FALSE;
+		cvswrite = 0;
 		break;
 	    case 'w':
-		cvswrite = TRUE;
+		cvswrite = 1;
 		break;
 	    case 't':
-		trace = TRUE;
+		trace = 1;
 		break;
 	    case 'n':
-		noexec = TRUE;
+		noexec = 1;
 	    case 'l':			/* Fall through */
-		logoff = TRUE;
+		logoff = 1;
 		break;
 	    case 'v':
+		/* Having the year here is a good idea, so people have
+		   some idea of how long ago their version of CVS was
+		   released.  */
 		(void) fputs (version_string, stdout);
 		(void) fputs (config_string, stdout);
 		(void) fputs ("\n", stdout);
-		(void) fputs ("Copyright (c) 1993-1994 Brian Berliner\n", stdout);
-		(void) fputs ("Copyright (c) 1993-1994 david d `zoo' zuhn\n", stdout);
-		(void) fputs ("Copyright (c) 1992, Brian Berliner and Jeff Polk\n", stdout);
-		(void) fputs ("Copyright (c) 1989-1992, Brian Berliner\n", stdout);
+		(void) fputs ("\
+Copyright (c) 1989-1998 Brian Berliner, david d `zoo' zuhn, \n\
+                        Jeff Polk, and other authors\n", stdout);
 		(void) fputs ("\n", stdout);
 		(void) fputs ("CVS may be copied only under the terms of the GNU General Public License,\n", stdout);
 		(void) fputs ("a copy of which can be found with the CVS distribution kit.\n", stdout);
+		(void) fputs ("\n", stdout);
+
+		(void) fputs ("Specify the --help option for further information about CVS\n", stdout);
+
 		exit (0);
 		break;
 	    case 'b':
-		Rcsbin = optarg;
-		rcsbin_update_env = 1;	/* need to update environment */
+		/* This option used to specify the directory for RCS
+		   executables.  But since we don't run them any more,
+		   this is a noop.  Silently ignore it so that .cvsrc
+		   and scripts and inetd.conf and such can work with
+		   either new or old CVS.  */
 		break;
 	    case 'T':
-		Tmpdir = optarg;
+		Tmpdir = xstrdup (optarg);
+		free_Tmpdir = 1;
 		tmpdir_update_env = 1;	/* need to update environment */
 		break;
 	    case 'e':
-		Editor = optarg;
+		Editor = xstrdup (optarg);
+		free_Editor = 1;
 		break;
 	    case 'd':
-		CVSroot = optarg;
+		CVSroot = xstrdup (optarg);
+		free_CVSroot = 1;
 		cvs_update_env = 1;	/* need to update environment */
 		break;
 	    case 'H':
 	        help = 1;
 		break;
             case 'f':
-		use_cvsrc = FALSE; /* unnecessary, since we've done it above */
+		use_cvsrc = 0; /* unnecessary, since we've done it above */
 		break;
 	    case 'z':
 #ifdef CLIENT_SUPPORT
@@ -497,6 +583,15 @@ main (argc, argv)
                    If no ENCRYPTION, we still accept -x, but issue an
                    error if we are being run as a client.  */
 		break;
+	    case 'a':
+#ifdef CLIENT_SUPPORT
+		cvsauthenticate = 1;
+#endif
+		/* If no CLIENT_SUPPORT, ignore -a, so that users can
+                   have it in their .cvsrc and not cause any trouble.
+                   We will issue an error later if stream
+                   authentication is not supported.  */
+		break;
 	    case '?':
 	    default:
                 usage (usg);
@@ -527,6 +622,9 @@ main (argc, argv)
     else
 	command_name = cm->fullname;	/* Global pointer for later use */
 
+    /* This should probably remain a warning, rather than an error,
+       for quite a while.  For one thing the version of VC distributed
+       with GNU emacs 19.34 invokes 'cvs rlog' instead of 'cvs log'.  */
     if (strcmp (argv[0], "rlog") == 0)
     {
 	error (0, 0, "warning: the rlog command is deprecated");
@@ -570,9 +668,15 @@ main (argc, argv)
 #endif /* HAVE_KERBEROS */
 
 
-#if defined(AUTH_SERVER_SUPPORT) && defined(SERVER_SUPPORT)
+#if (defined(AUTH_SERVER_SUPPORT) || defined (HAVE_GSSAPI)) && defined(SERVER_SUPPORT)
 	if (strcmp (command_name, "pserver") == 0)
 	{
+	    /* The reason that --allow-root is not a command option
+	       is mainly the comment in server() about how argc,argv
+	       might be from .cvsrc.  I'm not sure about that, and
+	       I'm not sure it is only true of command options, but
+	       it seems easier to make it a global option.  */
+
 	    /* Gets username and password from client, authenticates, then
 	       switches to run as that user and sends an ACK back to the
 	       client. */
@@ -581,15 +685,16 @@ main (argc, argv)
 	    /* Pretend we were invoked as a plain server.  */
 	    command_name = "server";
 	}
-#endif /* AUTH_SERVER_SUPPORT && SERVER_SUPPORT */
+#endif /* (AUTH_SERVER_SUPPORT || HAVE_GSSAPI) && SERVER_SUPPORT */
 
+#ifdef SERVER_SUPPORT
+	server_active = strcmp (command_name, "server") == 0;
 
 	/* Fiddling with CVSROOT doesn't make sense if we're running
            in server mode, since the client will send the repository
            directory after the connection is made. */
 
-#ifdef SERVER_SUPPORT
-	if (strcmp (command_name, "server") != 0)
+	if (!server_active)
 #endif
 	{
 	    char *CVSADM_Root;
@@ -603,13 +708,13 @@ main (argc, argv)
 	       ignores CVS directories and CVS/Root is likely to
 	       specify a different repository than the one we are
 	       importing to.  */
-#if 0
-	    if (lookup_command_attribute (command_name) & CVS_CMD_IGNORE_ADMROOT)
+
+	    if (lookup_command_attribute (command_name)
+                & CVS_CMD_IGNORE_ADMROOT)
+            {
 		CVSADM_Root = Name_Root((char *) NULL, (char *) NULL);
-#else
-	    if (strcmp (command_name, "import") != 0)
-		CVSADM_Root = Name_Root((char *) NULL, (char *) NULL);
-#endif
+            }
+
 	    if (CVSADM_Root != NULL)
 	    {
 		if (CVSroot == NULL || !cvs_update_env)
@@ -638,16 +743,12 @@ main (argc, argv)
 		       "cvs login" command.  Ahh, the things one
 		       discovers. */
 
-#if 0
-		    if (lookup_command_attribute (command_name) & CVS_CMD_USES_WORK_DIR)
-#else
-		    if ((strcmp (command_name, "checkout") != 0) &&
-			(strcmp (command_name, "login") != 0) &&
-			(strcmp (command_name, "rdiff") != 0) &&
-			(strcmp (command_name, "release") != 0) &&
-			(strcmp (command_name, "rtag") != 0))
-#endif
+		    if (lookup_command_attribute (command_name)
+                        & CVS_CMD_USES_WORK_DIR)
+                    {
 			need_to_create_root = 1;
+                    }
+
 		}
 	    }
 
@@ -691,9 +792,13 @@ main (argc, argv)
 
 	    if (!client_active)
 	    {
-		char path[PATH_MAX];
+		char *path;
 		int save_errno;
 
+		path = xmalloc (strlen (CVSroot_directory)
+				+ sizeof (CVSROOTADM)
+				+ 20
+				+ sizeof (CVSROOTADM_HISTORY));
 		(void) sprintf (path, "%s/%s", CVSroot_directory, CVSROOTADM);
 		if (!isaccessible (path, R_OK | X_OK))
 		{
@@ -712,6 +817,7 @@ main (argc, argv)
 		    error (0, 0, "Sorry, you don't have read/write access to the history file");
 		    error (1, save_errno, "%s", path);
 		}
+		free (path);
 	    }
 
 #ifdef HAVE_PUTENV
@@ -734,31 +840,21 @@ main (argc, argv)
 	   and/or remote path, on the other hand I'm not sure whether
 	   it is worth the trouble.  */
 
-	CurDir = xmalloc (PATH_MAX);
 #ifdef SERVER_SUPPORT
-	if (strcmp (command_name, "server") == 0)
-	    strcpy (CurDir, "<remote>");
+	if (server_active)
+	    CurDir = xstrdup ("<remote>");
 	else
 #endif
 	{
-            if (!getwd (CurDir))
-		error (1, 0, "cannot get working directory: %s", CurDir);
+	    CurDir = xgetwd ();
+            if (CurDir == NULL)
+		error (1, errno, "cannot get working directory");
 	}
 
 	if (Tmpdir == NULL || Tmpdir[0] == '\0')
 	    Tmpdir = "/tmp";
 
 #ifdef HAVE_PUTENV
-	/* Now, see if we should update the environment with the
-           Rcsbin value */
-	if (rcsbin_update_env)
-	{
-	    char *env;
-	    env = xmalloc (strlen (RCSBIN_ENV) + strlen (Rcsbin) + 1 + 1);
-	    (void) sprintf (env, "%s=%s", RCSBIN_ENV, Rcsbin);
-	    (void) putenv (env);
-	    /* do not free env, as putenv has control of it */
-	}
 	if (tmpdir_update_env)
 	{
 	    char *env;
@@ -769,24 +865,6 @@ main (argc, argv)
 	}
 #endif
 
-	/*
-	 * If Rcsbin is set to something, make sure it is terminated with
-	 * a slash character.  If not, add one.
-	 */
-	if (*Rcsbin)
-	{
-	    int len = strlen (Rcsbin);
-	    char *rcsbin;
-
-	    if (Rcsbin[len - 1] != '/')
-	    {
-		rcsbin = Rcsbin;
-		Rcsbin = xmalloc (len + 2);	/* one for '/', one for NULL */
-		(void) strcpy (Rcsbin, rcsbin);
-		(void) strcat (Rcsbin, "/");
-	    }
-	}
-
 #ifndef DONT_USE_SIGNALS
 	/* make sure we clean up on error */
 #ifdef SIGHUP
@@ -813,32 +891,39 @@ main (argc, argv)
 
 	gethostname(hostname, sizeof (hostname));
 
-#ifdef HAVE_SETVBUF
-	/*
-	 * Make stdout line buffered, so 'tail -f' can monitor progress.
-	 * Patch creates too much output to monitor and it runs slowly.
-	 */
-#  ifndef KLUDGE_FOR_WNT_TESTSUITE
-
-	if (strcmp (cm->fullname, "patch"))
-#    ifdef BUFSIZ  /* traditional SysV chokes when size == 0 */
-	    (void) setvbuf (stdout, (char *) NULL, _IOLBF, BUFSIZ);
-#    else
-	    (void) setvbuf (stdout, (char *) NULL, _IOLBF, 0);
-#    endif
-
-#  else /* KLUDGE_FOR_WNT_TESTSUITE */
-
-	    (void) setvbuf (stdout, (char *) NULL, _IONBF, 0);
-	    (void) setvbuf (stderr, (char *) NULL, _IONBF, 0);
-
-#  endif /* KLUDGE_FOR_WNT_TESTSUITE */
-
-#endif /* HAVE_SETVBUF */
+#ifdef KLUDGE_FOR_WNT_TESTSUITE
+	/* Probably the need for this will go away at some point once
+	   we call fflush enough places (e.g. fflush (stdout) in
+	   cvs_outerr).  */
+	(void) setvbuf (stdout, (char *) NULL, _IONBF, 0);
+	(void) setvbuf (stderr, (char *) NULL, _IONBF, 0);
+#endif /* KLUDGE_FOR_WNT_TESTSUITE */
 
 	if (use_cvsrc)
 	    read_cvsrc (&argc, &argv, command_name);
 
+	/* Parse the CVSROOT/config file, but only for local.  For the
+	   server, we parse it after we know $CVSROOT.  For the
+	   client, it doesn't get parsed at all, obviously.  The
+	   presence of the parse_config call here is not mean to
+	   predetermine whether CVSROOT/config overrides things from
+	   read_cvsrc and other such places or vice versa.  That sort
+	   of thing probably needs more thought.  */
+	if (1
+#ifdef SERVER_SUPPORT
+	    && !server_active
+#endif
+#ifdef CLIENT_SUPPORT
+	    && !client_active
+#endif
+	    )
+	{
+	    /* If there was an error parsing the config file, parse_config
+	       already printed an error.  We keep going.  Why?  Because
+	       if we didn't, then there would be no way to check in a new
+	       CVSROOT/config file to fix the broken one!  */
+	    parse_config (CVSroot_directory);
+	}
     } /* end of stuff that gets done if the user DOESN'T ask for help */
 
     err = (*(cm->func)) (argc, argv);
@@ -847,21 +932,35 @@ main (argc, argv)
     {
 	/* Update the CVS/Root file.  We might want to do this in
 	   all directories that we recurse into, but currently we
-	   don't.  */
+	   don't.  Note that if there is an error writing the file,
+	   we give an error/warning.  This is so if users try to rewrite
+	   CVS/Root with the -d option (a documented feature), they will
+	   either succeed, or be told why it didn't work.  */
 	Create_Root (NULL, CVSroot);
     }
 
     Lock_Cleanup ();
 
+    free (program_path);
+    if (free_CVSroot)
+	free (CVSroot);
+    if (free_Editor)
+	free (Editor);
+    if (free_Tmpdir)
+	free (Tmpdir);
+    root_allow_free ();
+
 #ifdef SYSTEM_CLEANUP
     /* Hook for OS-specific behavior, for example socket subsystems on
        NT and OS2 or dealing with windows and arguments on Mac.  */
     SYSTEM_CLEANUP ();
 #endif
 
-    if (err)
-	return (EXIT_FAILURE);
-    return 0;
+    /* This is exit rather than return because apparently that keeps
+       some tools which check for memory leaks happier.  */
+    exit (err ? EXIT_FAILURE : 0);
+	/* Keep picky/stupid compilers (e.g. Visual C++ 5.0) happy.  */
+	return 0;
 }
 
 char *
@@ -870,17 +969,20 @@ Make_Date (rawdate)
 {
     struct tm *ftm;
     time_t unixtime;
-    char date[256];			/* XXX bigger than we'll ever need? */
+    char date[MAXDATELEN];
     char *ret;
 
     unixtime = get_date (rawdate, (struct timeb *) NULL);
     if (unixtime == (time_t) - 1)
 	error (1, 0, "Can't parse date/time: %s", rawdate);
-#ifdef HAVE_RCS5
+
     ftm = gmtime (&unixtime);
-#else
-    ftm = localtime (&unixtime);
-#endif
+    if (ftm == NULL)
+	/* This is a system, like VMS, where the system clock is in local
+	   time.  Hopefully using localtime here matches the "zero timezone"
+	   hack I added to get_date.  */
+	ftm = localtime (&unixtime);
+
     (void) sprintf (date, DATEFORM,
 		    ftm->tm_year + (ftm->tm_year < 100 ? 0 : 1900),
 		    ftm->tm_mon + 1, ftm->tm_mday, ftm->tm_hour,
@@ -896,5 +998,5 @@ usage (cpp)
     (void) fprintf (stderr, *cpp++, program_name, command_name);
     for (; *cpp; cpp++)
 	(void) fprintf (stderr, *cpp);
-    exit (EXIT_FAILURE);
+    error_exit ();
 }
