@@ -14,11 +14,7 @@
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   GNU General Public License for more details.  */
 
 #include "cvs.h"
 
@@ -758,7 +754,10 @@ fncmp (const char *n1, const char *n2)
    rather than "canonical form").  The idea is that files we create
    would get thusly munged, but CVS can cope with their names being
    different the same way that the NT port copes with it if the user
-   renames a file from "foo" to "FOO".  */
+   renames a file from "foo" to "FOO".
+
+   Alternately, this kind of handling could/should go into CVS_FOPEN
+   and friends (if we want to do it like the Mac port, anyway).  */
 void
 fnfold (char *filename)
 {
@@ -808,7 +807,7 @@ last_component (path)
 {
     char *last = strrchr (path, '/');
 
-    if (last)
+    if (last && (last != path))
         return last + 1;
     else
         return path;
