@@ -34,6 +34,10 @@
    I have neither the CD-ROM nor a CD-ROM drive to put it in.  */
 #define HAVE_UTIME_NULL 1
 
+/* On Windows NT, when a file is being watched, utime expects a file
+   to be writable */
+#define UTIME_EXPECTS_WRITABLE
+
 /* Define if on MINIX.  */
 /* Hah.  */
 #undef _MINIX
@@ -100,20 +104,6 @@
    that's always there.  */
 #undef SERVER_SUPPORT
 
-/* the path to the gnu diff program on your system  */
-/* We don't need this for CLIENT side.  */
-#undef DIFF
-
-/* the path to the gnu grep program on your system  */
-/* We don't need this for CLIENT side.  */
-#undef GREP
-
-/* The number of bytes in a int.  */
-#define SIZEOF_INT 4
-
-/* The number of bytes in a long.  */
-#define SIZEOF_LONG 4
-
 /* Define if you have the connect function.  */
 /* Not used?  */
 #define HAVE_CONNECT
@@ -139,15 +129,9 @@
 /* Define if you have the krb_get_err_text function.  */
 #undef HAVE_KRB_GET_ERR_TEXT
 
-/* Define if you have the mkfifo function.  */
-#undef HAVE_MKFIFO
-
 /* Define if you have the putenv function.  */
 #define HAVE_PUTENV 1
 
-/* Define if you have the setvbuf function.  */
-#define HAVE_SETVBUF 1
-
 /* Define if you have the sigaction function.  */
 #undef HAVE_SIGACTION
 
@@ -270,37 +254,18 @@ unsigned int sleep (unsigned int);
 #define popen _popen
 #define pclose _pclose
 
-/* Under Windows NT, filenames are case-insensitive, and both / and \
-   are path component separators.  */
-#define FOLD_FN_CHAR(c) (WNT_filename_classes[(unsigned char) (c)])
-extern unsigned char WNT_filename_classes[];
-#define FILENAMES_CASE_INSENSITIVE 1
+/* When writing binary data to stdout, we better set
+   stdout to binary mode using setmode.  */
+#define USE_SETMODE_STDOUT 1
 
-/* Is the character C a path name separator?  Under
-   Windows NT, you can use either / or \.  */
-#define ISDIRSEP(c) (FOLD_FN_CHAR(c) == '/')
+/* Diff also has an ifdef for setmode, and it is HAVE_SETMODE.  */
+#define HAVE_SETMODE 1
 
-/* Like strcmp, but with the appropriate tweaks for file names.
-   Under Windows NT, filenames are case-insensitive but case-preserving,
-   and both \ and / are path element separators.  */
-extern int fncmp (const char *n1, const char *n2);
-
-/* Fold characters in FILENAME to their canonical forms.  
-   If FOLD_FN_CHAR is not #defined, the system provides a default
-   definition for this.  */
-extern void fnfold (char *FILENAME);
-
-/* #define this if your system terminates lines in text files with
-   CRLF instead of plain LF, and your I/O functions automatically
-   translate between using LF in memory and CRLF on disk, unless you
-   specifically tell them not to.  */
-#define LINES_CRLF_TERMINATED 1
-
-/* Read data from INFILE, and copy it to OUTFILE. 
-   Open INFILE using INFLAGS, and OUTFILE using OUTFLAGS.
-   This is useful for converting between CRLF and LF line formats.  */
-extern void convert_file (char *INFILE,  int INFLAGS,
-			  char *OUTFILE, int OUTFLAGS);
+/* Diff needs us to define this.  I think it could always be
+   -1 for CVS, because we pass temporary files to diff, but
+   config.h seems like the easiest place to put this, so for
+   now we put it here.  */
+#define same_file(s,t) (-1)
 
 /* This is where old bits go to die under Windows NT.  */
 #define DEVNULL "nul"
@@ -327,6 +292,19 @@ extern void init_winsock();
    it because _open_osfhandle doesn't work.  */
 #define NO_SOCKET_TO_FD 1
 
+/* This tells the client that, in addition to needing to use
+   send()/recv() to do socket I/O, the error codes for send()/recv()
+   and other socket operations are not available through errno.
+   Instead, this macro should be used to obtain an error code. */
+#define SOCK_ERRNO (WSAGetLastError ())
+
+/* This tells the client that, in addition to needing to use
+   send()/recv() to do socket I/O, the error codes for send()/recv()
+   and other socket operations are not known to strerror.  Instead,
+   this macro should be used to convert the error codes to strings. */
+#define SOCK_STRERROR sock_strerror
+extern char *sock_strerror (int errnum);
+
 /* The internal rsh client uses sockets not file descriptors.  Note
    that as the code stands now, it often takes values from a SOCKET and
    puts them in an int.  This is ugly but it seems like sizeof
