@@ -11,11 +11,7 @@
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   GNU General Public License for more details.  */
 
 /* These functions were moved out of subr.c because they need different
    definitions under operating systems (like, say, Windows NT) with different
@@ -28,9 +24,13 @@
 
 static int deep_remove_dir PROTO((const char *path));
 
-/*
- * Copies "from" to "to".
- */
+/* Copies "from" to "to".  Note that the functionality here is similar
+   to the win32 function CopyFile, but (1) we copy LastAccessTime and
+   CopyFile doesn't, (2) we set file attributes to the default set by
+   the C library and CopyFile copies them.  Neither #1 nor #2 was intentional
+   as far as I know, but changing them could be confusing, unless there
+   is some reason they should be changed (this would need more
+   investigation).  */
 void
 copy_file (from, to)
     const char *from;
@@ -408,33 +408,6 @@ rename_file (from, to)
 	error (1, errno, "cannot rename file %s to %s", from, to);
 }
 
-/* Windows NT doesn't have hard links or symbolic links.
-   There was only one place in CVS which used this function,
-   so I rewrote it to work another way, so this function isn't
-   used any more.  */
-#if 0
-/*
- * link a file, if possible.
- */
-int
-link_file (from, to)
-    const char *from;
-    const char *to;
-{
-    if (trace)
-#ifdef SERVER_SUPPORT
-	(void) fprintf (stderr, "%c-> link(%s,%s)\n",
-			(server_active) ? 'S' : ' ', from, to);
-#else
-	(void) fprintf (stderr, "-> link(%s,%s)\n", from, to);
-#endif
-    if (noexec)
-	return (0);
-
-    return (link (from, to));
-}
-#endif
-
 /*
  * unlink a file, if possible.
  */
@@ -660,103 +633,6 @@ xcmp (file1, file2)
     return (ret);
 }
 
-
-/* The equivalence class mapping for filenames.
-   Windows NT filenames are case-insensitive, but case-preserving.
-   Both / and \ are path element separators.
-   Thus, this table maps both upper and lower case to lower case, and
-   both / and \ to /.  */
-
-#if 0
-main ()
-{
-  int c;
-
-  for (c = 0; c < 256; c++)
-    {
-      int t;
-
-      if (c == '\\')
-        t = '/';
-      else
-        t = tolower (c);
-      
-      if ((c & 0x7) == 0x0)
-         printf ("    ");
-      printf ("0x%02x,", t);
-      if ((c & 0x7) == 0x7)
-         putchar ('\n');
-      else if ((c & 0x7) == 0x3)
-         putchar (' ');
-    }
-}
-#endif
-
-
-unsigned char
-WNT_filename_classes[] =
-{
-    0x00,0x01,0x02,0x03, 0x04,0x05,0x06,0x07,
-    0x08,0x09,0x0a,0x0b, 0x0c,0x0d,0x0e,0x0f,
-    0x10,0x11,0x12,0x13, 0x14,0x15,0x16,0x17,
-    0x18,0x19,0x1a,0x1b, 0x1c,0x1d,0x1e,0x1f,
-    0x20,0x21,0x22,0x23, 0x24,0x25,0x26,0x27,
-    0x28,0x29,0x2a,0x2b, 0x2c,0x2d,0x2e,0x2f,
-    0x30,0x31,0x32,0x33, 0x34,0x35,0x36,0x37,
-    0x38,0x39,0x3a,0x3b, 0x3c,0x3d,0x3e,0x3f,
-    0x40,0x61,0x62,0x63, 0x64,0x65,0x66,0x67,
-    0x68,0x69,0x6a,0x6b, 0x6c,0x6d,0x6e,0x6f,
-    0x70,0x71,0x72,0x73, 0x74,0x75,0x76,0x77,
-    0x78,0x79,0x7a,0x5b, 0x2f,0x5d,0x5e,0x5f,
-    0x60,0x61,0x62,0x63, 0x64,0x65,0x66,0x67,
-    0x68,0x69,0x6a,0x6b, 0x6c,0x6d,0x6e,0x6f,
-    0x70,0x71,0x72,0x73, 0x74,0x75,0x76,0x77,
-    0x78,0x79,0x7a,0x7b, 0x7c,0x7d,0x7e,0x7f,
-    0x80,0x81,0x82,0x83, 0x84,0x85,0x86,0x87,
-    0x88,0x89,0x8a,0x8b, 0x8c,0x8d,0x8e,0x8f,
-    0x90,0x91,0x92,0x93, 0x94,0x95,0x96,0x97,
-    0x98,0x99,0x9a,0x9b, 0x9c,0x9d,0x9e,0x9f,
-    0xa0,0xa1,0xa2,0xa3, 0xa4,0xa5,0xa6,0xa7,
-    0xa8,0xa9,0xaa,0xab, 0xac,0xad,0xae,0xaf,
-    0xb0,0xb1,0xb2,0xb3, 0xb4,0xb5,0xb6,0xb7,
-    0xb8,0xb9,0xba,0xbb, 0xbc,0xbd,0xbe,0xbf,
-    0xc0,0xc1,0xc2,0xc3, 0xc4,0xc5,0xc6,0xc7,
-    0xc8,0xc9,0xca,0xcb, 0xcc,0xcd,0xce,0xcf,
-    0xd0,0xd1,0xd2,0xd3, 0xd4,0xd5,0xd6,0xd7,
-    0xd8,0xd9,0xda,0xdb, 0xdc,0xdd,0xde,0xdf,
-    0xe0,0xe1,0xe2,0xe3, 0xe4,0xe5,0xe6,0xe7,
-    0xe8,0xe9,0xea,0xeb, 0xec,0xed,0xee,0xef,
-    0xf0,0xf1,0xf2,0xf3, 0xf4,0xf5,0xf6,0xf7,
-    0xf8,0xf9,0xfa,0xfb, 0xfc,0xfd,0xfe,0xff,
-};
-
-/* Like strcmp, but with the appropriate tweaks for file names.
-   Under Windows NT, filenames are case-insensitive but case-preserving,
-   and both \ and / are path element separators.  */
-int
-fncmp (const char *n1, const char *n2)
-{
-    while (*n1 && *n2
-           && (WNT_filename_classes[(unsigned char) *n1]
-	       == WNT_filename_classes[(unsigned char) *n2]))
-        n1++, n2++;
-    return (WNT_filename_classes[(unsigned char) *n1]
-            - WNT_filename_classes[(unsigned char) *n2]);
-}
-
-/* Fold characters in FILENAME to their canonical forms.  
-   If FOLD_FN_CHAR is not #defined, the system provides a default
-   definition for this.  */
-void
-fnfold (char *filename)
-{
-    while (*filename)
-    {
-        *filename = FOLD_FN_CHAR (*filename);
-	filename++;
-    }
-}
-
 
 /* Generate a unique temporary filename.  Returns a pointer to a newly
    malloc'd string containing the name.  Returns successfully or not at
@@ -778,6 +654,12 @@ int
 isabsolute (filename)
     const char *filename;
 {
+    /* FIXME: This routine seems to interact poorly with
+       strip_trailing_slashes.  For example, specify ":local:r:\" as
+       CVSROOT.  The CVS/Root file will contain ":local:r:" and then
+       isabsolute will complain about the root not being an absolute
+       pathname.  My guess is that strip_trailing_slashes is the right
+       place to fix this.  */
     return (ISDIRSEP (filename[0])
             || (filename[0] != '\0'
                 && filename[1] == ':'
@@ -795,67 +677,54 @@ last_component (char *path)
         if (ISDIRSEP (*scan))
 	    last = scan;
 
-    if (last)
+    if (last && (last != path))
         return last + 1;
     else
         return path;
 }
 
 
-/* Read data from INFILE, and copy it to OUTFILE. 
-   Open INFILE using INFLAGS, and OUTFILE using OUTFLAGS.
-   This is useful for converting between CRLF and LF line formats.  */
-void
-convert_file (char *infile,  int inflags,
-	      char *outfile, int outflags)
-{
-    int infd, outfd;
-    char buf[8192];
-    int len;
-
-    if ((infd = open (infile, inflags)) < 0)
-        error (1, errno, "couldn't read %s", infile);
-    if ((outfd = open (outfile, outflags, S_IWRITE)) < 0)
-        error (1, errno, "couldn't write %s", outfile);
-
-    while ((len = read (infd, buf, sizeof (buf))) > 0)
-        if (write (outfd, buf, len) < 0)
-	    error (1, errno, "error writing %s", outfile);
-    if (len < 0)
-        error (1, errno, "error reading %s", infile);
-
-    if (close (outfd) < 0)
-        error (0, errno, "warning: couldn't close %s", outfile);
-    if (close (infd) < 0)
-        error (0, errno, "warning: couldn't close %s", infile);
-}
-
-
 /* NT has two evironment variables, HOMEPATH and HOMEDRIVE, which,
    when combined as ${HOMEDRIVE}${HOMEPATH}, give the unix equivalent
    of HOME.  Some NT users are just too unixy, though, and set the
    HOME variable themselves.  Therefore, we check for HOME first, and
-   then try to combine the other two if that fails.  */
+   then try to combine the other two if that fails.
+
+   Looking for HOME strikes me as bogus, particularly if the only reason
+   is to cater to "unixy users".  On the other hand, if the reasoning is
+   there should be a single variable, rather than requiring people to
+   set both HOMEDRIVE and HOMEPATH, then it starts to make a little more
+   sense.
+
+   Win95: The system doesn't set HOME, HOMEDRIVE, or HOMEPATH (at
+   least if you set it up as the "all users under one user ID" or
+   whatever the name of that option is).  Based on thing overheard on
+   the net, it seems that users of the pserver client have gotten in
+   the habit of setting HOME (if you don't use pserver, you can
+   probably get away without having a reasonable return from
+   get_homedir.  Of course you lose .cvsrc and .cvsignore, but many
+   users won't notice).  So it would seem that we should be somewhat
+   careful if we try to change the current behavior.
+
+   NT 3.51 or NT 4.0: I haven't checked this myself, but I am told
+   that HOME gets set, but not to the user's home directory.  It is
+   said to be set to c:\users\default by default.  */
 
 char *
 get_homedir ()
 {
-    static char pathbuf[PATH_MAX * 2];
+    static char *pathbuf;
     char *hd, *hp;
 
-    if ((hd = getenv ("HOME")))
+    if (pathbuf != NULL)
+	return pathbuf;
+    else if ((hd = getenv ("HOME")))
 	return hd;
     else if ((hd = getenv ("HOMEDRIVE")) && (hp = getenv ("HOMEPATH")))
     {
-	/* Watch for buffer overruns. */
-
-#define cvs_min(x,y) ((x <= y) ? (x) : (y))
-
-	int ld = cvs_min (PATH_MAX, strlen (hd));
-	int lp = cvs_min (PATH_MAX, strlen (hp));
-
-	strncpy (pathbuf, hd, ld);
-	strncpy (pathbuf + ld, hp, lp);
+	pathbuf = xmalloc (strlen (hd) + strlen (hp) + 5);
+	strcpy (pathbuf, hd);
+	strcat (pathbuf, hp);
 
 	return pathbuf;
     }
@@ -893,6 +762,14 @@ expand_wild (argc, argv, pargc, pargv)
 	char *last_forw_slash, *last_back_slash, *end_of_dirname;
 	int dirname_length = 0;
 
+	/* FIXME: If argv[i] is ".", this code will expand it to the
+	   name of the current directory in its parent directory which
+	   will cause start_recursion to do all manner of strange things
+	   with it (culminating in an error).  This breaks "cvs co .".
+	   As nearly as I can guess, this bug has existed since
+	   expand_wild was first created.  At least, it is in CVS 1.9 (I
+	   just tried it).  */
+
 	/* FindFirstFile doesn't return pathnames, so we have to do
 	   this ourselves.  Luckily, it's no big deal, since globbing
 	   characters under Win32s can only occur in the last segment
@@ -908,6 +785,9 @@ expand_wild (argc, argv, pargc, pargv)
 
 #define cvs_max(x,y) ((x >= y) ? (x) : (y))
 
+	/* FIXME: this comparing a NULL pointer to a non-NULL one is
+	   extremely ugly, and I strongly suspect *NOT* sanctioned by
+	   ANSI C.  The code should just use last_component instead.  */
 	end_of_dirname = cvs_max (last_forw_slash, last_back_slash);
 
 	if (end_of_dirname == NULL)
@@ -956,7 +836,16 @@ expand_wild (argc, argv, pargc, pargv)
 
 		/* Copy the file name. */
 		
-		strcat (new_argv[new_argc], fdata.cFileName);
+		if (fncmp (argv[i] + dirname_length, fdata.cFileName) == 0)
+		    /* We didn't expand a wildcard; we just matched a filename.
+		       Use the file name as specified rather than the filename
+		       which exists in the directory (they may differ in case).
+		       This is needed to make cvs add on a directory consistently
+		       use the name specified on the command line, but it is
+		       probably a good idea in other contexts too.  */
+		    strcpy (new_argv[new_argc], argv[i]);
+		else
+		    strcat (new_argv[new_argc], fdata.cFileName);
 
 		new_argc++;
 
